import io
import logging
import os
import re
import time
from datetime import date, datetime, timedelta

import boto3
import pandas as pd

from mycroft.common import aws_s3, gen_signature, reputation
from mycroft.common.aws_s3 import get_s3_client, upload_dataframe_to_s3
from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)

"""
Query athena to get hash org id in reputation table
"""


def get_hash_org_df(aws_data, database, reputation_table, force_query):
    query = f"""
    SELECT value1 AS SHA256_HASH, orgid, env
    FROM {database}.{reputation_table}
    where type = 'SHA256'
    and authlisttype = 'WHITE_LIST' and not regexp_like(value1, '^[0]*$')
    """
    return aws_data.get_dataframe(
        gen_signature(query, "policy-automation"), "hash_orgid.csv", force_query
    )


"""
Query athena to get org device count in device_group and policy table
"""


def get_org_device_count_df(
    aws_data, database, policy_device_mapping_table, policy_table, force_query
):
    limitation = "LIMIT 100000" if database == "cdca" else ""
    query = f"""
    SELECT t2.org_id AS orgid, t.env AS env, t.device_count AS device_count
    FROM (
        SELECT pd.env AS env,
               pd.device_group_id AS device_group_id,
               count(distinct pd.device_id) AS device_count
               FROM {database}.{policy_device_mapping_table}
               AS pd group by pd.device_group_id, pd.env {limitation}) AS t
    LEFT JOIN (
        SELECT distinct p.env AS env,
        p.org_id AS org_id,
        p.device_group_id AS device_group_id FROM {database}.{policy_table} AS p {limitation}) AS t2
      ON t.env = t2.env
      AND t.device_group_id = t2.device_group_id
    WHERE t2.org_id IS NOT NULL;
    """
    return aws_data.get_dataframe(
        gen_signature(query, "policy-automation"), "org_device_count.csv", force_query
    )


"""
Query athena to get cert org id in reputation table
"""


def get_cert_org_df(aws_data, database, reputation_table, force_query):
    query = (
        "SELECT value1 AS cert_signedby, value2 AS cert_authority, orgid, env \
    FROM {}.{} \
    where authlisttype = 'WHITE_LIST' "
        "and type = 'CERT' "
        "order by orgid".format(database, reputation_table)
    )
    return aws_data.get_dataframe(
        gen_signature(query, "policy-automation"), "cert_orgid.csv", force_query
    )


"""
function to join cert/hash org id with org device count
"""


def calc_device_org_count(
    cert_hash_org_df, org_device_count_df, aggregate_name, total_device_count
):
    cert_hash_device_count_df = cert_hash_org_df.merge(
        org_device_count_df, on=["orgid", "env"], how="left"
    ).drop("env", axis=1)
    cert_hash_device_count_df["device_count"].fillna(0, inplace=True)
    cert_hash_device_sum_df = (
        cert_hash_device_count_df.groupby(aggregate_name)
        .agg({"device_count": "sum", "orgid": "size"})
        .reset_index()
        .sort_values(by=["device_count"], ascending=False)
    )
    cert_hash_device_sum_df.rename(columns={"orgid": "org_count"}, inplace=True)
    cert_hash_device_sum_df["device_count_percent"] = (
        cert_hash_device_sum_df["device_count"] / total_device_count
    )
    cert_hash_device_sum_df = cert_hash_device_sum_df.reset_index().drop(
        ["index"], axis=1
    )
    return cert_hash_device_sum_df


"""
function to generate hash/cert org_count and device count
"""


def generate_cert_hash_basic_prevalent(
    aws_data,
    database,
    reputation_table,
    policy_table,
    policy_device_mapping_table,
    generation_type,
    force_query,
    region,
):
    logger.info("generating org device count dataframe")
    org_device_count_df = get_org_device_count_df(
        aws_data, database, policy_device_mapping_table, policy_table, force_query
    )
    total_device_count = org_device_count_df["device_count"].sum()
    logging.info("total device count: %d", int(total_device_count))
    if generation_type == "cert":
        logging.info("generating cert org id dataframe")
        cert_org_df = get_cert_org_df(aws_data, database, reputation_table, force_query)
        cert_org_df["cert_authority"].replace("", "*")
        logger.info("generate cert device org count dataframe")
        cert_device_org_count_df = calc_device_org_count(
            cert_org_df,
            org_device_count_df,
            ["cert_signedby", "cert_authority"],
            total_device_count,
        )
        cert_device_org_count_df["region"] = region
        return cert_device_org_count_df, cert_org_df
    else:
        hash_org_df = get_hash_org_df(aws_data, database, reputation_table, force_query)
        logger.info("generate hash device org count dataframe")
        hash_device_org_count_df = calc_device_org_count(
            hash_org_df, org_device_count_df, "SHA256_HASH", total_device_count
        )
        hash_device_org_count_df["region"] = region
        return hash_device_org_count_df, hash_org_df


"""
function to find all hashes whose reputation are NOT_LISTED,
RESOLVING or ADAPTIVE_WHITE
"""


def find_candidate_hashes(df):
    excluding_start_time = time.time()
    all_hash_list = list(df["SHA256_HASH"])
    logger.info("Size of hash list in reputation: %d", len(all_hash_list))
    n = 100
    chunked_hash_lists = [
        all_hash_list[i * n : (i + 1) * n]
        for i in range((len(all_hash_list) + n - 1) // n)
    ]
    index = 0
    all_candidate_hashes = []
    while index < len(chunked_hash_lists):
        try:
            hash_list = chunked_hash_lists[index]
            logger.debug(f"Querying chunked hash list: {hash_list}")
            results = reputation.process_hashes(hash_list)
            candidate_hashes = []
            if "NOT_LISTED" in results:
                candidate_hashes = list(results["NOT_LISTED"])
                logger.debug("not_listed hash size: %d", len(results["NOT_LISTED"]))
            if "ADAPTIVE_WHITE_LIST" in results:
                candidate_hashes = candidate_hashes + list(
                    results["ADAPTIVE_WHITE_LIST"]
                )
                logger.debug(
                    "adaptive_whitelist hash size: %d",
                    len(results["ADAPTIVE_WHITE_LIST"]),
                )
            all_candidate_hashes = all_candidate_hashes + candidate_hashes
            row_len = df.shape[0]
            logger.debug(
                "index: %d, hash list size: %d, candidate hashes in the batch size: %d, all candidate hashes size: %d",
                index,
                row_len,
                len(candidate_hashes),
                len(all_candidate_hashes),
            )
            index += 1
        except Exception as e:
            logger.error("Error for hash list index: %d, error: %s", index, str(e))
            logger.error("sleeping")
            time.sleep(1)
    if len(all_candidate_hashes) > 0:
        df = df[df["SHA256_HASH"].isin(all_candidate_hashes)]
        df = df.sort_values(by=["org_count"], ascending=False)
        logger.debug(
            f"Excluding white list time: {(time.time() - excluding_start_time)}"
        )
        return df
    else:
        logger.info("No hashes in the hash pool!")
        logger.debug(
            f"Excluding white list time: {(time.time() - excluding_start_time)}"
        )
        return None


# TODO: Need to add linux/mac support for this
"""
Query Athena to get the most common proc path for hashes in the hash_list
"""


def get_most_common_path(
    aws_data,
    start_time,
    end_time,
    chunked_hash_list,
    intermediate_event_db,
    intermediate_event_table,
):
    try:
        hash_list = ",".join("'" + str(h) + "'" for h in chunked_hash_list)
        query = f"""
        SELECT temp.proc_sha256hash AS SHA256_HASH,
               temp.proc_path AS proc_path FROM
                (SELECT proc_sha256hash,
                        lower(proc_path) AS proc_path,
                        ROW_Number() OVER(PARTITION BY proc_sha256hash
                                          ORDER BY CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) desc)
                                         AS device_count_rank
                FROM {intermediate_event_db}.{intermediate_event_table}
                WHERE event_date BETWEEN '{start_time}' and '{end_time}'
                AND proc_sha256hash IN ({hash_list}) AND proc_path != '' AND dev_type='WINDOWS'
                GROUP BY 1,2) temp
            WHERE temp.device_count_rank = 1
        """
        df = aws_data.get_dataframe(
            gen_signature(query, "policy-automation"),
            "hash_proc_device_count.csv",
            True,
        )
        if df is None:
            logger.debug("FAILED in executing query: %s", query)
            logger.debug("Divide the length by half")
            length = len(chunked_hash_list) // 2
            chunked_hash_sublists = [
                chunked_hash_list[:length],
                chunked_hash_list[length:],
            ]
            dfs = []
            for chunked_hash_sublist in chunked_hash_sublists:
                dfs.append(
                    get_most_common_path(
                        aws_data,
                        start_time,
                        end_time,
                        chunked_hash_sublist,
                        intermediate_event_db,
                        intermediate_event_table,
                    )
                )
            return pd.concat(dfs, axis=0)
        else:
            return df
    except Exception as e:
        logger.debug("Error in the most common path query, exception: %s", str(e))


"""
function to extract file name from a file path
"""


def extract_file_name(file_path):
    try:
        if file_path is None or len(file_path) == 0:
            return ""
        basename = re.search(r"[^\\/]+(?=[\\/]?$)", file_path)
        if basename:
            return basename.group(0)
    except Exception as e:
        logger.error(
            "Error in extract file name. Error details: %s. Proc path: %s",
            str(e),
            str(file_path),
        )
        return ""


"""
function to generate most common path and file name for all hashes in the hash_df
spit the hash list into chunks, each chunk has size 2000
query athena to get the most common path in the chunk
concatenate all query results to a dataframe
"""


def generate_most_common_path_df(
    aws_data,
    start_time,
    end_time,
    hash_df,
    intermediate_event_db,
    intermediate_event_table,
):
    generate_most_common_path_start_time = time.time()
    all_hash_list = hash_df["SHA256_HASH"].tolist()
    n = 2000
    chunked_hash_lists = [
        all_hash_list[i * n : (i + 1) * n]
        for i in range((len(all_hash_list) + n - 1) // n)
    ]
    logger.debug("Total batches: %d", len(chunked_hash_lists))
    hash_proc_device_count_df_list = []
    index = 1
    for chunked_hash_list in chunked_hash_lists:
        logger.debug(
            "round: %d, chunked_hash_list_len: %d", index, len(chunked_hash_list)
        )
        temp_df = get_most_common_path(
            aws_data,
            start_time,
            end_time,
            chunked_hash_list,
            intermediate_event_db,
            intermediate_event_table,
        )
        logger.debug("hash size: %d", temp_df.shape[0])
        hash_proc_device_count_df_list.append(temp_df)
        index += 1
    hash_proc_device_count_df = pd.concat(
        hash_proc_device_count_df_list, ignore_index=True
    )
    logging.debug(
        f"hash_df size: {hash_df.shape[0]}, hash_proc_device_count_df size: {hash_proc_device_count_df.shape[0]}"
    )
    all_hash_df = pd.merge(
        hash_df, hash_proc_device_count_df, on="SHA256_HASH", how="left"
    )
    all_hash_df["proc_path"].fillna("", inplace=True)
    all_hash_df["file_name"] = all_hash_df["proc_path"].apply(extract_file_name)
    # make all proc_path as empty string for sanitization purpose
    all_hash_df["proc_path"] = ""
    logger.debug(
        f"Generate most common path time: {(time.time() - generate_most_common_path_start_time)}"
    )
    return all_hash_df


"""
Main function to call
Refer to README.md for more info about this function
"""


def generate_cert_hash_prevalent_data(
    region,
    database,
    reputation_table,
    policy_table,
    policy_device_mapping_table,
    intermediate_event_db,
    intermediate_event_table,
    query_output_bucket,
    query_output_folder,
    query_output_location,
    csv_output_list_location,
    existing_rule_list_location,
    end_time,
    generation_type="proc_sha256hash",
    prevalence_threshold=1,
    force_query=True,
    cross_account_arn=None,
):
    try:
        start_time_obj = datetime.strptime(end_time, "%Y-%m-%d") - timedelta(days=29)
        start_time = datetime.strftime(start_time_obj, "%Y-%m-%d")
        params = {
            "region": region,
            "database": database,
            "bucket": query_output_bucket,
            "outputBucket": query_output_bucket,
            "outputFolder": query_output_folder,
            "outputLocation": query_output_location,
        }
        aws_data = aws_s3.AwsS3(params, cross_account_arn=cross_account_arn)
        if generation_type == "cert":
            cert_device_org_count_df, cert_org_df = generate_cert_hash_basic_prevalent(
                aws_data,
                database,
                reputation_table,
                policy_table,
                policy_device_mapping_table,
                generation_type,
                force_query,
                region,
            )
            cert_device_org_count_df = cert_device_org_count_df[
                cert_device_org_count_df["org_count"] > prevalence_threshold
            ]
            cert_device_org_count_df = cert_device_org_count_df[
                cert_device_org_count_df["cert_authority"] != "*"
            ]
            client = get_s3_client(cross_account_arn=cross_account_arn)
            upload_dataframe_to_s3(
                cert_device_org_count_df, csv_output_list_location, client
            )
            # cert_device_org_count_df.to_csv(csv_output_list_location, index=False)
            cert_org_df = cert_org_df.rename(columns={"orgid": "org_id"})
            upload_dataframe_to_s3(cert_org_df, existing_rule_list_location, client)
            # cert_org_df.to_csv(existing_rule_list_location, index=False)
            return (csv_output_list_location, existing_rule_list_location)
        elif generation_type == "proc_sha256hash":
            hash_device_org_count_df, hash_org_df = generate_cert_hash_basic_prevalent(
                aws_data,
                database,
                reputation_table,
                policy_table,
                policy_device_mapping_table,
                generation_type,
                force_query,
                region,
            )
            hash_device_org_count_df["org_count"].fillna(0, inplace=True)
            hash_device_org_count_df = hash_device_org_count_df[
                hash_device_org_count_df["org_count"] > prevalence_threshold
            ]
            filtered_hash_org_count_df = find_candidate_hashes(hash_device_org_count_df)
            if filtered_hash_org_count_df is not None:
                logging.info(
                    f"After finding candidate hashes. Hash pool size: {filtered_hash_org_count_df.shape[0]}"
                )
                filtered_hash_org_count_common_path_df = generate_most_common_path_df(
                    aws_data,
                    start_time,
                    end_time,
                    filtered_hash_org_count_df,
                    intermediate_event_db,
                    intermediate_event_table,
                )
                filtered_hash_org_count_common_path_df = (
                    filtered_hash_org_count_common_path_df.sort_values(
                        by=["org_count"], ascending=False
                    )
                )
                logger.info(
                    f"[generate_cert_hash_prevalent_data] Prepare write dataframe to CSV: {csv_output_list_location}"
                )
                client = get_s3_client(cross_account_arn=cross_account_arn)
                upload_dataframe_to_s3(
                    filtered_hash_org_count_common_path_df,
                    csv_output_list_location,
                    client,
                )
                # filtered_hash_org_count_common_path_df.to_csv(
                #    csv_output_list_location, index=False
                # )
                logger.debug(
                    f"[generate_cert_hash_prevalent_data] Prepare write dataframe to CSV: {existing_rule_list_location}"
                )
                hash_org_df = hash_org_df.rename(
                    columns={"orgid": "org_id", "SHA256_HASH": "proc_sha256hash"}
                )
                upload_dataframe_to_s3(hash_org_df, existing_rule_list_location, client)
                # hash_org_df.to_csv(existing_rule_list_location, index=False)
                return (csv_output_list_location, existing_rule_list_location)
            else:
                return None, None
        else:
            logger.error(
                "generation_type error. Should be either 'cert' or 'proc_sha256hash'!"
            )
            return None
    except Exception as e:
        logger.error(
            "Error in generating cert hash prevalent data. Details: %s", str(e)
        )
        return None
