#!/usr/bin/env python
# coding: utf-8
import json
import os
import tempfile
import time
import traceback

import awswrangler as wr
import boto3
import numpy as np
import pandas as pd
import psycopg2
import requests
from botocore.exceptions import ClientError
from dotenv import load_dotenv
from wavefront_sdk.client_factory import WavefrontClientFactory  # type: ignore
from wavefront_sdk.common import metric_to_line_data  # type: ignore

from mycroft.common import aws_s3
from mycroft.common.aws_s3 import (
    assume_role_session,
    get_s3_client,
    get_s3_resource,
    upload_dataframe_to_s3,
)
from mycroft.common.logging import get_logger
from mycroft.common.org_num_normalizer import org_num_normalizer
from mycroft.common.pathtree.aggregator import group_perm_rules_path
from mycroft.common.pathtree.path_tree import PathTree
from mycroft.generate_prevalence_data.generate_reputation_hash_cert_prevalent import (
    generate_cert_hash_prevalent_data,
)
from mycroft.generate_suggestions.get_reasons import (
    gen_daily_events_with_impact_number,
    gen_monthly_events_with_impact_number,
)
from mycroft.generate_suggestions.get_suggestion_pool import (
    gen_params_for_it_tools_reputation_pool,
    gen_permission_rules_pool,
    get_existing_it_tools_rules,
)
from mycroft.generate_suggestions.impact_score import generate_impact_score
from mycroft.generate_suggestions.write_suggestions_to_api import (
    create_perm_recomm_json,
    create_reputation_recomm_json,
)
from psc_orgkey import decode_orgkey, encode_orgkey

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)

# Load and process config/env variables
load_dotenv()

bad_rule_paths = ["**\\*", "*\\*", "c:\\*", "c:\\**"]

# ENVS = os.getenv("ENVS", "prod01,prod02,prod05,eap01").split(",")
# logger.debug(f"ENVS: {ENVS}")

PREVALENCE_THRESH = int(os.getenv("PREVALENCE_THRESH", 1))
logger.debug(f"PREVALENCE THRESHOLD: {PREVALENCE_THRESH}")

GROUP_PREVALENCE_THRESH = int(os.getenv("GROUP_PREVALENCE_THRESH", 5))
logger.debug(f"GROUP PREVALENCE THRESHOLD: {GROUP_PREVALENCE_THRESH}")

REMOVE_PERM_RULES_START_DBL_STR = eval(
    os.getenv("REMOVE_PERM_RULES_START_DBL_STR", "True")
)
logger.debug(f"REMOVE_PERM_RULES_START_DBL_STR: {REMOVE_PERM_RULES_START_DBL_STR}")

RULES_POOL_BUCKET = os.getenv(
    "RULES_POOL_BUCKET", "cdca-prod-us-east-1-intermediate-data"
)
logger.debug(f"RULES_POOL_BUCKET: {RULES_POOL_BUCKET}")

EVENTS_POOL_BUCKET = os.getenv(
    "EVENTS_POOL_BUCKET", "cdca-prod-us-east-1-intermediate-data"
)
logger.debug(f"EVENTS_POOL_BUCKET: {EVENTS_POOL_BUCKET}")

NORMALIZER_BUCKET = os.getenv(
    "NORMALIZER_BUCKET", "cdca-prod-us-east-1-intermediate-data"
)
logger.debug(f"EVENTS_POOL_BUCKET: {EVENTS_POOL_BUCKET}")

REGION = os.getenv("REGION", "us-east-1")
logger.debug(f"REGION: {REGION}")

STAGE = os.getenv("STAGE", "dev")
logger.debug(f"STAGE: {STAGE}")

FLINK_PRIMARY_INFRA_DB_NAME = os.getenv("FLINK_PRIMARY_INFRA_DB_NAME", "cdca")
logger.debug(f"FLINK_PRIMARY_INFRA_DB_NAME: {FLINK_PRIMARY_INFRA_DB_NAME}")

BASELINE_POLICY_VIEW_NAME = os.getenv("BASELINE_POLICY_VIEW_NAME", "policy_view")
logger.debug(f"BASELINE_POLICY_VIEW_NAME: {BASELINE_POLICY_VIEW_NAME}")

BASELINE_REPUTATION_TABLE_NAME = os.getenv(
    "BASELINE_REPUTATION_TABLE_NAME", "policy_org_reputation_override"
)
logger.debug(f"BASELINE_REPUTATION_TABLENAME: {BASELINE_REPUTATION_TABLE_NAME}")

BASELINE_DEVICE_MAPPING_TABLE = os.getenv(
    "BASELINE_DEVICE_MAPPING_TABLE", "policy_device_mapping"
)
logger.debug(f"BASELINE_DEVICE_MAPPING_TABLE: {BASELINE_DEVICE_MAPPING_TABLE}")

QUERY_OUTPUT_BUCKET = os.getenv(
    "QUERY_OUTPUT_BUCKET", "cdca-policy-suggest-analytics-dev-athena-outputs"
)
logger.debug(f"QUERY_OUTPUT_BUCKET: {QUERY_OUTPUT_BUCKET}")

DAILY_INTERMEDIATE_DB = os.getenv("DAILY_INTERMEDIATE_DB")
logger.debug(f"DAILY_INTERMEDIATE_DB: {DAILY_INTERMEDIATE_DB}")

DAILY_INTERMEDIATE_TABLE = os.getenv("DAILY_INTERMEDIATE_TABLE")
logger.debug(f"DAILY_INTERMEDIATE_TABLE: {DAILY_INTERMEDIATE_TABLE}")

INTERMEDIATE_DB = os.getenv("INTERMEDIATE_DB")
logger.debug(f"INTERMEDIATE_DB: {INTERMEDIATE_DB}")

INTERMEDIATE_DB_HOST = os.getenv("INTERMEDIATE_DB_HOST", "intermediate_db_host")
logger.debug(f"INTERMEDIATE_DB_HOST: {INTERMEDIATE_DB_HOST}")

INTERMEDIATE_DB_PORT = os.getenv("INTERMEDIATE_DB_PORT", "intermediate_db_port")
logger.debug(f"INTERMEDIATE_DB_PORT: {INTERMEDIATE_DB_PORT}")

INTERMEDIATE_DB_USER = os.getenv("INTERMEDIATE_DB_USER", "intermediate_db_user")
logger.debug(f"INTERMEDIATE_DB_USER: {INTERMEDIATE_DB_USER}")

INTERMEDIATE_DB_PASSWORD = os.getenv(
    "INTERMEDIATE_DB_PASSWORD", "intermediate_db_password"
)

INTERNAL_INGRESS_URL = os.getenv("INTERNAL_INGRESS_URL", "endpoint_url")
logger.debug(f"INTERNAL_INGRESS_URL: {INTERNAL_INGRESS_URL}")

API_KEY = os.getenv("API_KEY", "api_key")
CONNECTOR_ID = os.getenv("CONNECTOR_ID", "connector_id")

ALERT_CNT_THRESHOLD = os.getenv("ALERT_CNT_THRESHOLD", 1)
logger.debug(f"ALERT_CNT_THRESHOLD: {ALERT_CNT_THRESHOLD}")

DEVICE_CNT_THRESHOLD = os.getenv("DEVICE_CNT_THRESHOLD", 5)
logger.debug(f"DEVICE_CNT_THRESHOLD: {DEVICE_CNT_THRESHOLD}")

RDS_SECRET_NAME = os.getenv("RDS_SECRET_NAME", "prod/policy-automation/rds")
logger.debug(f"RDS_SECRET_NAME: {RDS_SECRET_NAME}")

URL_SECRET_NAME = os.getenv("URL_SECRET_NAME", "prod/policy-automation/url")
logger.debug(f"URL_SECRET_NAME: {URL_SECRET_NAME}")

SERVICE_TOKEN_SECRET_NAME = os.getenv(
    "SERVICE_TOKEN_SECRET_NAME", "prod/policy-suggestion/service-tokens"
)
logger.debug(f"SERVICE_TOKEN_SECRET_NAME: {SERVICE_TOKEN_SECRET_NAME}")

RECOMMENDATION_NUM_PER_ORG = os.getenv("RECOMMENDATION_NUM_PER_ORG", 50)
logger.debug(f"RECOMMENDATION_NUM_PER_ORG: {RECOMMENDATION_NUM_PER_ORG}")

CROSS_ACCOUNT_ARN = os.getenv("MOSAIC1_IAM_ARN", None)


def aws_get_secrets(secret_name, secret_region):
    # Create a Secrets Manager client
    session = boto3.session.Session()

    client = session.client(
        service_name="secretsmanager",
        region_name=secret_region,
    )
    try:
        # Fetch the secret from secret manager
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
        secrets = json.loads(get_secret_value_response["SecretString"])
        return secrets
    except ClientError as e:
        error_msg = f"Error in reading secret. Details:{e}"
        raise Exception(error_msg)


def aws_get_db_parms(secret_name, secret_region):
    """Read the value from aws secret manager services"""
    try:
        secrets = aws_get_secrets(secret_name, secret_region)
        rds_host = secrets["host"]
        rds_port = secrets["port"]
        rds_username = secrets["username"]
        rds_password = secrets["password"]
        return rds_host, rds_port, rds_username, rds_password
    except Exception as e:
        error_msg = f"Error in getting db params. Details:{e}"
        raise Exception(error_msg)


def aws_get_db_parms_mosaic2(secret_name, secret_region):
    """Read the value from aws secret manager services"""

    try:
        secrets = aws_get_secrets(secret_name, secret_region)
        rds_host = secrets["host"]
        rds_port = secrets["port"]
        rds_username = secrets["username"]
        rds_password = secrets["password"]
        return rds_host, rds_port, rds_username, rds_password, ""
    except Exception as e:
        error_msg = f"Error in getting db params. Details:{e}"
        raise Exception(error_msg)


def env_get_db_parms():
    """Read the value from environment variables"""
    return (
        INTERMEDIATE_DB_HOST,
        INTERMEDIATE_DB_PORT,
        INTERMEDIATE_DB_USER,
        INTERMEDIATE_DB_PASSWORD,
        # Required in mosaicv2 INTERMEDIATE_DB,
    )


def env_get_db_parms_mosaic2():
    """Read the value from environment variables"""
    return (
        INTERMEDIATE_DB_HOST,
        INTERMEDIATE_DB_PORT,
        INTERMEDIATE_DB_USER,
        INTERMEDIATE_DB_PASSWORD,
        INTERMEDIATE_DB,
    )


def env_get_urls_tokens():
    return INTERNAL_INGRESS_URL, SERVICE_TOKEN


def get_db_parms_mosiac2(secret_name, secret_region):
    # Attempt to get from env vars
    host, port, user, pwd, dbname = env_get_db_parms_mosaic2()

    # If host was found, assume we're all set.
    # If not, try reading from aws secret manager.
    if not host:
        host, port, user, pwd, dbname = aws_get_db_parms_mosaic2(
            secret_name, secret_region
        )

    return host, port, user, pwd, dbname


def get_db_parms(secret_name, secret_region):
    # Attempt to get from env vars
    host, port, user, pwd = env_get_db_parms()

    # If host was found, assume we're all set.
    # If not, try reading from aws secret manager.
    if not host:
        host, port, user, pwd = aws_get_db_parms(secret_name, secret_region)

    return host, port, user, pwd


def k8s_get_token(api_key, connector_id):
    return f"{api_key}/{connector_id}"


"""MosaicV2
RDS_DB, RDS_PORT, RDS_USERNAME, RDS_PWD, RDS_DB_NAME = get_db_parms(RDS_SECRET_NAME, REGION)
logger.debug(f"RDS_DB: {RDS_DB}, RDS_PORT: {RDS_PORT}, RDS_USERNAME: {RDS_USERNAME}, RDS_DB_NAME: {RDS_DB_NAME}")
"""

RDS_DB, RDS_PORT, RDS_USERNAME, RDS_PWD = get_db_parms(RDS_SECRET_NAME, REGION)
logger.debug(f"RDS_DB: {RDS_DB}, RDS_PORT: {RDS_PORT}, RDS_USERNAME: {RDS_USERNAME}")

SERVICE_TOKEN = k8s_get_token(API_KEY, CONNECTOR_ID)
logger.debug(f"SERVICE_TOKEN: {SERVICE_TOKEN}")


class invalidStageError(Exception):
    pass


def create_rds_tables():
    # create recommendations table
    create_recommendations_command = """CREATE TABLE IF NOT EXISTS recommendations (
                                        recommendation_type varchar(50) NOT NULL,
                                        value1 TEXT NOT NULL,
                                        value2 TEXT DEFAULT '',
                                        value3 TEXT DEFAULT '',
                                        env varchar(50) NOT NULL,
                                        org_id INT NOT NULL,
                                        policy_id INT,
                                        dev_cnt INT NOT NULL,
                                        org_cnt INT NOT NULL,
                                        alert_cnt INT NOT NULL,
                                        alert_dev_cnt INT NOT NULL,
                                        comment1 TEXT,
                                        comment2 TEXT,
                                        impact_score real DEFAULT 0.0,
                                        CONSTRAINT perm_rule_unique_rule UNIQUE(recommendation_type, env, org_id, policy_id, value1, value2, value3)
                                    )"""
    run_command_on_rds(create_recommendations_command)

    # clear all recommendations
    clear_recommendations_command = """delete from recommendations"""
    run_command_on_rds(clear_recommendations_command)

    # Create hash events table
    create_hash_events_command = """CREATE TABLE IF NOT EXISTS hash_events (
                                    env varchar(50),
                                    org_id INT,
                                    proc_sha256hash TEXT,
                                    dev_cnt INT,
                                    alert_cnt INT
                                )"""
    run_command_on_rds(create_hash_events_command)

    # clear all hash_events
    clear_hash_events = """delete from hash_events"""
    run_command_on_rds(clear_hash_events)

    # create hash pool table
    create_hash_pool_command = """CREATE TABLE IF NOT EXISTS hash_pool (
                                    proc_sha256hash TEXT,
                                    device_count real,
                                    org_count real,
                                    dev_count_per real,
                                    region varchar(20),
                                    proc_path TEXT,
                                    file_name TEXT
                                )"""
    run_command_on_rds(create_hash_pool_command)

    # clear all hash_pool
    clear_hash_pool = """delete from hash_pool"""
    run_command_on_rds(clear_hash_pool)

    # create existing_hash_rules
    create_existing_hash_rules = """CREATE TABLE IF NOT EXISTS existing_hash_rules (
                                        proc_sha256hash TEXT,
                                        org_id INT,
                                        env varchar(50)
                                    )"""
    run_command_on_rds(create_existing_hash_rules)

    # clear all existing cert rules
    clear_existing_hash_rules = """delete from existing_hash_rules"""
    run_command_on_rds(clear_existing_hash_rules)

    # create cert events table
    create_cert_events_command = """CREATE TABLE IF NOT EXISTS cert_events (
                                    env varchar(50),
                                    org_id INT,
                                    cert_signedby TEXT,
                                    cert_authority TEXT,
                                    dev_cnt INT,
                                    alert_cnt INT
                                )"""
    run_command_on_rds(create_cert_events_command)

    # clear all cert_events
    clear_cert_events = """delete from cert_events"""
    run_command_on_rds(clear_cert_events)

    # create cert pool table
    create_cert_pool_command = """CREATE TABLE IF NOT EXISTS cert_pool (
                                    cert_signedby TEXT,
                                    cert_authority TEXT,
                                    device_count real,
                                    org_count real,
                                    dev_count_per real,
                                    region varchar(20)
                                )"""
    run_command_on_rds(create_cert_pool_command)

    # clear all cert_pool
    clear_cert_pool = """delete from cert_pool"""
    run_command_on_rds(clear_cert_pool)

    # create existing_cert_rules
    create_existing_cert_rules = """CREATE TABLE IF NOT EXISTS existing_cert_rules (
                                    cert_signedby TEXT,
                                    cert_authority TEXT,
                                    org_id INT,
                                    env varchar(50)
                                )"""
    run_command_on_rds(create_existing_cert_rules)

    # clear all existing cert rules
    clear_existing_cert_rules = """delete from existing_cert_rules"""
    run_command_on_rds(clear_existing_cert_rules)

    # create existing_it_tools_rules
    create_existing_it_tools = """CREATE TABLE IF NOT EXISTS existing_it_tools_rules (
                                    rule_path TEXT,
                                    org_id INT,
                                    env varchar(50)
                                )"""
    run_command_on_rds(create_existing_it_tools)

    # clear all existing it_tools rules
    clear_existing_it_tools_rules = """delete from existing_it_tools_rules"""
    run_command_on_rds(clear_existing_it_tools_rules)


def run_command_on_rds(command):
    conn = psycopg2.connect(
        # Required in mosaicv2 dbname=RDS_DB_NAME,
        host=RDS_DB,
        port=RDS_PORT,
        user=RDS_USERNAME,
        password=RDS_PWD,
    )
    cur = conn.cursor()
    try:
        cur.execute(command)
        conn.commit()
    except Exception as e:
        logger.error(f"Error in run command on RDS. Error details: {e}")
    cur.close()


def query_fetch_on_rds(command):
    conn = psycopg2.connect(
        # Required in mosaicv2 dbname=RDS_DB_NAME,
        host=RDS_DB,
        port=RDS_PORT,
        user=RDS_USERNAME,
        password=RDS_PWD,
    )
    cur = conn.cursor()
    res = None
    try:
        cur.execute(command)
        res = cur.fetchall()
        conn.commit()
    except Exception as e:
        logger.error(f"Error in run command on RDS. Error details: {e}")
    cur.close()
    return res


def run_copy_command_on_rds(local_file, command):
    logger.debug("[run_copy_command_on_rds] Prepare connection")
    conn = psycopg2.connect(
        # Required in mosaicv2 dbname=RDS_DB_NAME,
        host=RDS_DB,
        port=RDS_PORT,
        user=RDS_USERNAME,
        password=RDS_PWD,
    )
    logger.debug("[run_copy_command_on_rds] connection established")
    f = open(local_file)
    cur = conn.cursor()
    try:
        cur.copy_expert(command, f)
        conn.commit()
    except Exception as e:
        logger.error(f"Error in copy csv file to RDS. Error details: {e}")
    cur.close()


def download_dump_csv_into_table(
    s3_file_bucket, s3_file_key, table_name, cross_account_arn
):
    try:
        s3 = get_s3_resource(cross_account_arn=cross_account_arn)
        logger.info(f"Prepare downloading CSV. cross account: {cross_account_arn}")
        with tempfile.TemporaryDirectory() as temp_directory:
            temp_filename = f"{temp_directory}/temp.csv"
            s3.Bucket(s3_file_bucket).download_file(s3_file_key, temp_filename)

            copy_command = (
                f"""COPY {table_name} from stdin with csv header delimiter ',' """
            )
            run_copy_command_on_rds(temp_filename, copy_command)
    except Exception as e:
        logger.error(
            f"Error in downloading and dumping csv file into rds table. Error details: {e}"
        )
        raise Exception("Failed to download CSV and insert to RDS table")


def dump_local_file_into_rds(generation_type, local_output_file):
    # Dump local file into a temp database table
    try:
        start_time = time.time()
        if generation_type == "it_tools":
            command = (
                """COPY recommendations from stdin with delimiter ',' NULL AS '' """
            )
        else:
            command = """COPY recommendations from stdin with delimiter ',' """
        run_copy_command_on_rds(local_output_file, command)
        dump_temp_table_time = time.time() - start_time
        logger.debug(f"dump temp table time: {dump_temp_table_time}")
    except Exception as e:
        raise Exception(
            f"Failed in dumping rule recommendations into recommendations table. Details: {e}"
        )


# flake8: noqa: C901
def gen_proc_rules_sug(generate_date, generation_type="perm_rule", stage="bootstrap"):
    rules_pool_folder = f"{generate_date}/rules_pool/{generation_type}"
    rules_pool_filename = "rules_pool.csv"
    grouped_rules_pool_folder = f"{generate_date}/rules_pool/{generation_type}"
    grouped_rules_pool_filename = "grouped_rules_pool.csv"
    grouped_rules_path = f"s3://{RULES_POOL_BUCKET}/{grouped_rules_pool_folder}/{grouped_rules_pool_filename}"
    events_pool_folder = f"{generate_date}/events_pool/{generation_type}"

    if generation_type not in ("perm_rule", "it_tools"):
        raise Exception(
            "generation_type value error. It should be either 'perm_rule' or 'it_tools'!"
        )

    # generate permission rules pool or it tools pool based on the generation type
    try:
        start_time = time.time()
        if generation_type == "perm_rule":
            s3_path_to_rules_pool = gen_permission_rules_pool(
                prevalence_thresh=PREVALENCE_THRESH,
                s3_output_bucket=RULES_POOL_BUCKET,
                s3_output_folder=rules_pool_folder,
                s3_output_filename=rules_pool_filename,
                region=REGION,
                database=FLINK_PRIMARY_INFRA_DB_NAME,
                table=BASELINE_POLICY_VIEW_NAME,
                query_bucket=QUERY_OUTPUT_BUCKET,
            )
        else:
            s3_path_to_rules_pool = f"s3://{RULES_POOL_BUCKET}/static_it_tools_pool/static_it_tools_pool.csv"
            s3_key = get_existing_it_tools_rules(
                s3_output_bucket=RULES_POOL_BUCKET,
                s3_output_folder=f"{generate_date}/existing_rules/it_tools",
                s3_output_filename="it_tools_rule_list.csv",
                region=REGION,
                database=FLINK_PRIMARY_INFRA_DB_NAME,
                table=BASELINE_REPUTATION_TABLE_NAME,
                query_bucket=QUERY_OUTPUT_BUCKET,
                cross_account_arn=CROSS_ACCOUNT_ARN,
            )
        if s3_path_to_rules_pool is None:
            raise Exception(f"Failed in generating {generation_type} pool.")
        get_rules_pool_duration = time.time() - start_time
        logger.debug(
            f"path_to_perm_rules_rep: {s3_path_to_rules_pool} get_perm_rules_pool_duration: {get_rules_pool_duration}"
        )
    except Exception as e:
        logger.error(
            f"Failed in {generation_type} pool generation. Error details: {e} "
        )
        raise Exception(
            f"Failed in {generation_type} pool generation. Error details: {e} "
        )

    # Group (aggregate) permission rules suggestion generation pool
    try:
        start_time = time.time()
        if generation_type == "perm_rule":
            group_perm_rules_path(
                s3_path_to_rules_pool,
                grouped_rules_path,
                grp_prev_thresh=GROUP_PREVALENCE_THRESH,
                remove_start_dbl_str=REMOVE_PERM_RULES_START_DBL_STR,
                bad_rule_paths=bad_rule_paths,
            )
        else:
            gen_params_for_it_tools_reputation_pool(
                static_it_tools_rules_path=s3_path_to_rules_pool,
                s3_query_bucket=QUERY_OUTPUT_BUCKET,
                s3_output_filename=grouped_rules_path,
                region=REGION,
                database=FLINK_PRIMARY_INFRA_DB_NAME,
                table=BASELINE_REPUTATION_TABLE_NAME,
                query_bucket=QUERY_OUTPUT_BUCKET,
                cross_role_arn=CROSS_ACCOUNT_ARN,
            )
        rule_path_grouping_time = time.time() - start_time
        logger.debug(
            f"grouped_perm_rules_path: {grouped_rules_path} perm_rule_path_grouping_time: {rule_path_grouping_time}"
        )
    except Exception as e:
        raise Exception(f"Grouping {generation_type} rule paths for today failed. {e} ")

    try:
        start_time = time.time()
        if generation_type == "perm_rule":
            reason_subject = "proc_path"
        else:
            reason_subject = "it_tools"
        if stage == "bootstrap":
            (
                s3_path,
                events_bucket,
                events_file_key,
            ) = gen_monthly_events_with_impact_number(
                dev_type="WINDOWS",
                generate_date=generate_date,
                reason_subject=reason_subject,
                db_name=DAILY_INTERMEDIATE_DB,
                tbl_name=DAILY_INTERMEDIATE_TABLE,
                region=REGION,
                query_bucket=QUERY_OUTPUT_BUCKET,
                output_bucket=EVENTS_POOL_BUCKET,
                output_folder=events_pool_folder,
                alert_cnt_filter=ALERT_CNT_THRESHOLD,
                # dev01 requires lower filter to get more data
                dev_cnt_filter=4 if STAGE == "int" else DEVICE_CNT_THRESHOLD,
                cross_account_arn=CROSS_ACCOUNT_ARN,
            )
        elif stage == "daily_run":
            (
                s3_path,
                events_bucket,
                events_file_key,
            ) = gen_daily_events_with_impact_number(
                dev_type="WINDOWS",
                generate_date=generate_date,
                reason_subject=reason_subject,
                db_name=DAILY_INTERMEDIATE_DB,
                tbl_name=DAILY_INTERMEDIATE_TABLE,
                region=REGION,
                output_bucket=EVENTS_POOL_BUCKET,
                output_folder=events_pool_folder,
                alert_cnt_filter=ALERT_CNT_THRESHOLD,
                dev_cnt_filter=DEVICE_CNT_THRESHOLD,
            )
        else:
            logger.error(
                f"An invalid stage was provided when generating hash recommendations: {stage}"
            )
            raise invalidStageError()
        events_query_time = time.time() - start_time
        logger.debug(
            f"events_bucket: {events_bucket}, events_file_key: {events_file_key}, events_query_time: {events_query_time}"
        )
    except Exception as e:
        raise Exception("Query monthly events failed. {0} ".format(e))

    # build the rule path tree on the grouped rule paths
    try:
        start_time = time.time()
        # Build the rule path tree from the IT tools suggestion pool
        built_path_tree = PathTree("root")
        built_path_tree.add_rules_to_tree(grouped_rules_path, CROSS_ACCOUNT_ARN)
        perm_rule_tree_build_time = time.time() - start_time
        logger.debug(f"perm_rule_tree_build_time: {perm_rule_tree_build_time}")
    except Exception as e:
        raise Exception("Building permission rules path-tree failed {0} ".format(e))

    # global cache for storing matched results
    rule_path_match = dict()
    logger.info(
        "IT_TOOL: Preparing for generating recommendation.csv file in rules_pool folder"
    )

    # Match the proc_path to the path_tree
    try:
        with tempfile.TemporaryDirectory() as temp_directory:
            local_temp_file = f"{temp_directory}/all_events.csv"
            local_output_file = f"{temp_directory}/{generate_date}_{generation_type}_recommendations.csv"
            start_time = time.time()
            (num_it_tools_suggestions) = built_path_tree.find_multi_proc_path_in_tree(
                local_output_file,
                local_temp_file,
                rule_path_match,
                generation_type,
                events_bucket,
                events_file_key,
                CROSS_ACCOUNT_ARN,
            )
            rule_tree_matching_time = time.time() - start_time
            logger.info(
                f"local temp file stores all rule recommendations: {local_temp_file}, num of generated rule suggestion: {num_it_tools_suggestions}, permission rules matching duration {rule_tree_matching_time}"
            )
            if num_it_tools_suggestions > 0:
                logger.info(
                    f"save s3 file with recommendations to bucket: {RULES_POOL_BUCKET}, key: {rules_pool_folder}/recommendations.csv"
                )
                aws_data = aws_s3.AwsS3({}, cross_account_arn=CROSS_ACCOUNT_ARN)
                aws_data.upload_file(
                    local_output_file,
                    RULES_POOL_BUCKET,
                    f"{rules_pool_folder}/recommendations.csv",
                )
    except Exception as e:
        raise Exception(f"Permission rules proc-path to path tree matching failed {e} ")


def insert_it_tools_rec_intermediate_rds(
    generation_type, region, tools_bucket, it_tool_recommendation_key
):
    try:
        with tempfile.TemporaryDirectory() as temp_directory:
            local_temp_file = f"{temp_directory}/recommendations.csv"
            s3 = boto3.resource("s3", region_name=region)
            bucket = s3.Bucket(tools_bucket)
            objs = list(bucket.objects.filter(Prefix=it_tool_recommendation_key))
            if len(objs) > 0 and objs[0].key == it_tool_recommendation_key:
                bucket.download_file(it_tool_recommendation_key, local_temp_file)
                dump_local_file_into_rds(generation_type, local_temp_file)
    except Exception as e:
        raise Exception(f"Permission rules proc-path to path tree matching failed {e} ")


def gen_hash_sug(generate_date: str, stage: str):
    """
    params:
    """
    hash_pool_s3_bucket = RULES_POOL_BUCKET
    hash_pool_s3_key = f"{generate_date}/rules_pool/proc_sha256hash/hash_list_with_common_path_and_file_name.csv"
    hash_pool_s3_location = f"s3://{hash_pool_s3_bucket}/{hash_pool_s3_key}"
    existing_hash_rule_s3_key = (
        f"{generate_date}/existing_rules/proc_sha256hash/hash_rule_list.csv"
    )
    existing_hash_rules_s3_location = (
        f"s3://{hash_pool_s3_bucket}/{existing_hash_rule_s3_key}"
    )
    events_pool_folder = f"{generate_date}/events_pool/proc_sha256hash"
    try:
        start_time = time.time()
        if stage == "bootstrap":
            (
                hash_events_s3_path,
                hash_events_s3_bucket,
                hash_events_s3_key,
            ) = gen_monthly_events_with_impact_number(
                dev_type="WINDOWS",
                generate_date=generate_date,
                reason_subject="proc_sha256hash",
                db_name=DAILY_INTERMEDIATE_DB,
                tbl_name=DAILY_INTERMEDIATE_TABLE,
                region=REGION,
                query_bucket=QUERY_OUTPUT_BUCKET,
                output_bucket=EVENTS_POOL_BUCKET,
                output_folder=events_pool_folder,
                alert_cnt_filter=ALERT_CNT_THRESHOLD,
                dev_cnt_filter=DEVICE_CNT_THRESHOLD,
                cross_account_arn=CROSS_ACCOUNT_ARN,
            )
        elif stage == "daily_run":
            (
                hash_events_s3_path,
                hash_events_s3_bucket,
                hash_events_s3_key,
            ) = gen_daily_events_with_impact_number(
                dev_type="WINDOWS",
                generate_date=generate_date,
                reason_subject="proc_sha256hash",
                db_name=DAILY_INTERMEDIATE_DB,
                tbl_name=DAILY_INTERMEDIATE_TABLE,
                region=REGION,
                query_bucket=QUERY_OUTPUT_BUCKET,
                output_bucket=EVENTS_POOL_BUCKET,
                output_folder=events_pool_folder,
                alert_cnt_filter=ALERT_CNT_THRESHOLD,
                dev_cnt_filter=DEVICE_CNT_THRESHOLD,
            )

        else:
            logger.error(
                f"An invalid stage was provided when generating hash recommendations: {stage}"
            )
            raise invalidStageError()

        get_events_duration = time.time() - start_time
        logger.debug(
            f"path_to_events: {hash_events_s3_path} gen_events_duration: {get_events_duration}"
        )
        if hash_events_s3_path is None:
            raise Exception("Error: hash_events_s3_path is None")
    except Exception as e:
        raise Exception(f"Get events failed. {e}")

    try:
        start_time = time.time()
        logger.info(f"START: hash generate_cert_hash_prevalent_data ")
        (
            path_to_hash_rep,
            path_to_existing_hash_list,
        ) = generate_cert_hash_prevalent_data(
            region=REGION,
            database=FLINK_PRIMARY_INFRA_DB_NAME,
            reputation_table=BASELINE_REPUTATION_TABLE_NAME,
            policy_table=BASELINE_POLICY_VIEW_NAME,
            policy_device_mapping_table=BASELINE_DEVICE_MAPPING_TABLE,
            intermediate_event_db=DAILY_INTERMEDIATE_DB,
            intermediate_event_table=DAILY_INTERMEDIATE_TABLE,
            query_output_bucket=QUERY_OUTPUT_BUCKET,
            query_output_folder="",
            query_output_location=f"s3://{QUERY_OUTPUT_BUCKET}/",
            csv_output_list_location=hash_pool_s3_location,
            existing_rule_list_location=existing_hash_rules_s3_location,
            end_time=generate_date,
            generation_type="proc_sha256hash",
            cross_account_arn=CROSS_ACCOUNT_ARN,
        )
        get_hash_pool_duration = time.time() - start_time
        if path_to_hash_rep is not None:
            logger.info(
                f"END: hash generate_cert_hash_prevalent_data "
                f"path_to_hash_pool: {path_to_hash_rep}, existing_hash_rule_list: {path_to_existing_hash_list}, "
                f"get_hash_pool_duration: {get_hash_pool_duration}"
            )
        else:
            logger.info(
                f"END: hash generate_cert_hash_prevalent_data "
                f"No hash found in the hash pool. get_hash_pool_duration: {get_hash_pool_duration}"
                f"Create empty dataframe csv"
            )
            columns = [
                "SHA256_HASH",
                "device_count,org_count",
                "device_count_percent",
                "region",
                "proc_path",
                "file_name",
            ]
            df = pd.DataFrame(columns=columns)
            client = get_s3_client(cross_account_arn=CROSS_ACCOUNT_ARN)
            upload_dataframe_to_s3(df, hash_pool_s3_location, client)
    except Exception as e:
        traceback.print_exc()
        raise Exception(
            "Get hash reputation suggestion pool for today failed. {0} ".format(e)
        )


# insert only the hash pool data into rds
def insert_s3_hash_pool_intermediate_rds(
    s3_bucket, hash_pool_s3_key, cross_account_arn
):
    # Download and dump hash_pool only into the RDS tables.
    try:
        start_time = time.time()
        download_dump_csv_into_table(
            s3_bucket, hash_pool_s3_key, "hash_pool", cross_account_arn
        )

        download_and_dump_duration = time.time() - start_time
        logger.debug(f"download_and_dump_duration: {download_and_dump_duration}")

    except Exception as e:
        raise Exception(f"Download and dump hash events and pool tables failed. {e}")


def gen_cert_sug(generate_date, stage):
    cert_pool_s3_bucket = RULES_POOL_BUCKET
    cert_pool_s3_key = f"{generate_date}/rules_pool/cert/cert_list.csv"
    existing_cert_rules_s3_key = (
        f"{generate_date}/existing_rules/cert/cert_rule_list.csv"
    )
    cert_pool_s3_location = f"s3://{cert_pool_s3_bucket}/{cert_pool_s3_key}"
    existing_cert_pool_s3_location = (
        f"s3://{cert_pool_s3_bucket}/{existing_cert_rules_s3_key}"
    )
    events_pool_folder = f"{generate_date}/events_pool/cert"
    try:
        start_time = time.time()
        if stage == "bootstrap":
            (
                cert_events_s3_path,
                cert_events_s3_bucket,
                cert_events_s3_key,
            ) = gen_monthly_events_with_impact_number(
                dev_type="WINDOWS",
                generate_date=generate_date,
                reason_subject="cert",
                db_name=DAILY_INTERMEDIATE_DB,
                tbl_name=DAILY_INTERMEDIATE_TABLE,
                region=REGION,
                query_bucket=QUERY_OUTPUT_BUCKET,
                output_bucket=EVENTS_POOL_BUCKET,
                output_folder=events_pool_folder,
                alert_cnt_filter=ALERT_CNT_THRESHOLD,
                dev_cnt_filter=DEVICE_CNT_THRESHOLD,
                cross_account_arn=CROSS_ACCOUNT_ARN,
            )
        elif stage == "daily_run":
            (
                cert_events_s3_path,
                cert_events_s3_bucket,
                cert_events_s3_key,
            ) = gen_daily_events_with_impact_number(
                dev_type="WINDOWS",
                generate_date=generate_date,
                reason_subject="cert",
                db_name=DAILY_INTERMEDIATE_DB,
                tbl_name=DAILY_INTERMEDIATE_TABLE,
                region=REGION,
                query_bucket=QUERY_OUTPUT_BUCKET,
                output_bucket=EVENTS_POOL_BUCKET,
                output_folder=events_pool_folder,
                alert_cnt_filter=ALERT_CNT_THRESHOLD,
                dev_cnt_filter=DEVICE_CNT_THRESHOLD,
            )

        else:
            logger.error(
                f"An invalid stage was provided when generating cert recommendations: {stage}"
            )
            raise invalidStageError()

        get_events_duration = time.time() - start_time
        logger.debug(
            f"path_to_events: {cert_events_s3_path} gen_events_duration: {get_events_duration}"
        )

    except Exception as e:
        raise Exception(f"Get events failed. {e}")
    try:
        start_time = time.time()
        (
            path_to_cert_rep,
            path_to_existing_cert_list,
        ) = generate_cert_hash_prevalent_data(
            region=REGION,
            database=FLINK_PRIMARY_INFRA_DB_NAME,
            reputation_table=BASELINE_REPUTATION_TABLE_NAME,
            policy_table=BASELINE_POLICY_VIEW_NAME,
            policy_device_mapping_table=BASELINE_DEVICE_MAPPING_TABLE,
            intermediate_event_db=DAILY_INTERMEDIATE_DB,
            intermediate_event_table=DAILY_INTERMEDIATE_TABLE,
            query_output_bucket=QUERY_OUTPUT_BUCKET,
            query_output_folder="",
            query_output_location=f"s3://{QUERY_OUTPUT_BUCKET}/",
            csv_output_list_location=cert_pool_s3_location,
            existing_rule_list_location=existing_cert_pool_s3_location,
            end_time=generate_date,
            generation_type="cert",
            cross_account_arn=CROSS_ACCOUNT_ARN,
        )
        get_cert_pool_duration = time.time() - start_time
        logger.debug(
            f"path_to_cert_rep: {path_to_cert_rep}, existing_cert_rule_list: {path_to_existing_cert_list}, get_cert_pool_duration: {get_cert_pool_duration}"
        )
    except Exception as e:
        raise Exception(f"Get cert reputation suggestion pool for today failed. {e} ")


# insert only the cert pool data into rds
def insert_s3_cert_pool_intermediate_rds(
    s3_bucket, cert_pool_s3_key, cross_account_arn
):
    try:
        start_time = time.time()
        download_dump_csv_into_table(
            s3_bucket, cert_pool_s3_key, "cert_pool", cross_account_arn
        )

        download_and_dump_duration = time.time() - start_time
        logger.debug(f"download_and_dump_duration: {download_and_dump_duration}")

    except Exception as e:
        raise Exception(f"Download and dump cert events and pool tables failed. {e}")


# export the pool data from RDS into csv file
def export_pool_data(s3_bucket_list, hash_key, cert_key):
    # combine hash/cert data from all regions in RDS and export as csv files to each region
    """Required SSL settings for mosaic2
    ssl_ctx = ssl.create_default_context()
    ssl_ctx.check_hostname = False
    ssl_ctx.verify_mode = ssl.CERT_REQUIRED

    conn = wr.postgresql.pg8000.Connection(
        host=RDS_DB,
        port=RDS_PORT,
        database=RDS_DB_NAME,
        user=RDS_USERNAME,
        password=RDS_PWD,
        ssl_context=ssl_ctx,
    )
    """
    try:
        logger.info(f"[export_pool_data] START: s3_buckets=[{s3_bucket_list}]")
        conn = wr.postgresql.pg8000.Connection(
            host=RDS_DB,
            port=RDS_PORT,
            database="postgres",
            user=RDS_USERNAME,
            password=RDS_PWD,
        )
        logger.info("[export_pool_data] postgres connection established")

        hash_query = """SELECT p2.proc_sha256hash, p2.device_count, p2.org_count, p2.dev_count_per, p.proc_path, p.file_name
            FROM
                (SELECT proc_sha256hash,
                        SUM(device_count) as device_count,
                        SUM(org_count) as org_count,
                        SUM(dev_count_per) as dev_count_per
                        FROM hash_pool GROUP BY proc_sha256hash) p2
            LEFT JOIN hash_pool p
                ON p.proc_sha256hash = p2.proc_sha256hash
                AND p.org_count =
                    (SELECT MAX(p3.org_count)
                        FROM hash_pool p3
                        WHERE p3.proc_sha256hash = p.proc_sha256hash
                        AND coalesce(p3.file_name, '') <> '')
            WHERE p.proc_sha256hash is not null AND coalesce(p.file_name, '') <> '';
        """
        cert_query = """SELECT cert_signedby,
                               cert_authority,
                               SUM(device_count) as device_count,
                               SUM(org_count) as org_count,
                               SUM(dev_count_per) as dev_count_per
                        FROM cert_pool
                        GROUP BY cert_signedby, cert_authority;
        """

        # generate the dataframes from the queries
        hash_df = wr.postgresql.read_sql_query(sql=hash_query, con=conn)
        cert_df = wr.postgresql.read_sql_query(sql=cert_query, con=conn)

        # close the connection
        logger.info(f"[export_pool_data] postgres connection closed")
        conn.close()

        """Write results as csv file and save it into regional s3
        This might fail if previous query results are empty.
        """
        for s3_bucket in s3_bucket_list:
            if "cdca" not in s3_bucket:
                # mosaic2 buckets
                wr.s3.to_csv(
                    df=hash_df,
                    path=f"s3://{s3_bucket}/{hash_key}",
                )
                wr.s3.to_csv(
                    df=cert_df,
                    path=f"s3://{s3_bucket}/{cert_key}",
                )
                logger.info(
                    f"Complete writing query results to Pop s3://{s3_bucket}/{hash_key} and  s3://{s3_bucket}/{cert_key}"
                )
            else:
                # mosaic1 buckets, need cross account session
                session = assume_role_session(
                    CROSS_ACCOUNT_ARN, "ExportPoolDataSession"
                )
                wr.s3.to_csv(
                    df=hash_df,
                    path=f"s3://{s3_bucket}/{hash_key}",
                    boto3_session=session,
                )
                wr.s3.to_csv(
                    df=cert_df,
                    path=f"s3://{s3_bucket}/{cert_key}",
                    boto3_session=session,
                )
                logger.info(
                    f"Complete writing query results to s3://{s3_bucket}/{hash_key} and  s3://{s3_bucket}/{cert_key}"
                )
    except ClientError as client_error:
        traceback.print_exc()
        logger.error(
            f"[export_pool_data] Failed to write to s3. Skip for now. {client_error}"
        )
        pass
    except Exception as e:
        traceback.print_exc()
        raise Exception(f"[export_pool_data] Failed {e}")


# merge the cert events with existing rules to filter only new recommendations, then join events with the pool data
def merge_hash(
    aws_data,
    s3_bucket,
    hash_events_s3_key,
    hash_pool_s3_key,
    existing_hash_rules_s3_key,
):
    try:
        logger.info("START: merge_hash")
        events_df = aws_data.get_dataframe_from_s3(s3_bucket, hash_events_s3_key)
        events_df = get_org_id_env(events_df)
        pool_df = aws_data.get_dataframe_from_s3(s3_bucket, hash_pool_s3_key)
        existing_df = aws_data.get_dataframe_from_s3(
            s3_bucket, existing_hash_rules_s3_key
        )
        # set indexes
        events_df.set_index(["proc_sha256hash", "org_id", "env"], inplace=True)
        pool_df.set_index(["proc_sha256hash"], inplace=True)
        existing_df["exists"] = 1
        existing_df.set_index(["proc_sha256hash", "org_id", "env"], inplace=True)
        # join on existing
        events_df = events_df.join(existing_df, on=["proc_sha256hash", "org_id", "env"])
        # remove items that are existing
        events_df = events_df[events_df["exists"].isna()]
        # join on pool
        events_df = events_df.join(pool_df, on=["proc_sha256hash"])
        events_df = events_df[events_df["org_count"].notnull()]
        events_df = events_df.drop(
            columns=["exists", "device_count", "dev_count_per", "Unnamed: 0"]
        ).reset_index()
        # rename columns to correct recommendation fields and drop orgKey
        events_df.drop("org_key", axis=1, inplace=True, errors="ignore")
        events_df.columns = [
            "value1",
            "org_id",
            "env",
            "dev_cnt",
            "alert_cnt",
            "org_cnt",
            "comment1",
            "comment2",
        ]
        # add missing columns
        events_df["recommendation_type"] = "hash"
        events_df["alert_dev_cnt"] = 0
        logger.info("END: merge_hash")
        return events_df
    except Exception as e:
        logger.debug(traceback.format_exc())
        raise Exception(f"Failed in merge_hash. Error details: {e}")


# merge the cert events with existing rules to filter only new recommendations, then join events with the pool data
def merge_cert(
    aws_data,
    s3_bucket,
    cert_events_s3_key,
    cert_pool_s3_key,
    existing_cert_rules_s3_key,
):
    try:
        logger.info("START: merge_cert")
        events_df = aws_data.get_dataframe_from_s3(s3_bucket, cert_events_s3_key)
        events_df = get_org_id_env(events_df)
        pool_df = aws_data.get_dataframe_from_s3(s3_bucket, cert_pool_s3_key)
        existing_df = aws_data.get_dataframe_from_s3(
            s3_bucket, existing_cert_rules_s3_key
        )
        # set indexes
        events_df["cert_signedby"] = events_df["cert_signedby"].astype(object)
        events_df["cert_authority"] = events_df["cert_authority"].astype(object)
        events_df.set_index(
            ["cert_signedby", "cert_authority", "org_id", "env"], inplace=True
        )
        pool_df.set_index(["cert_signedby", "cert_authority"], inplace=True)
        existing_df["exists"] = 1
        existing_df.set_index(
            ["cert_signedby", "cert_authority", "org_id", "env"], inplace=True
        )
        # join on existing
        events_df = events_df.join(
            existing_df, on=["cert_signedby", "cert_authority", "org_id", "env"]
        )
        # remove items that are existing
        events_df = events_df[events_df["exists"].isna()]
        # join on pool
        events_df = events_df.join(pool_df, on=["cert_signedby", "cert_authority"])
        events_df = events_df[events_df["org_count"].notnull()]
        events_df = events_df.drop(
            columns=["exists", "device_count", "dev_count_per", "Unnamed: 0"]
        ).reset_index()
        # rename columns to correct recommendation fields and drop org_key
        events_df.drop("org_key", axis=1, inplace=True, errors="ignore")
        events_df.columns = [
            "value1",
            "value2",
            "org_id",
            "env",
            "dev_cnt",
            "alert_cnt",
            "org_cnt",
        ]
        # add missing fields
        events_df["recommendation_type"] = "cert"
        events_df["alert_dev_cnt"] = 0
        logger.info("END: merge_cert")
        return events_df
    except Exception as e:
        logger.debug(traceback.format_exc())
        raise Exception(f"Failed in merge_cert. Error details: {e}")


# merge the it_tool recommendations with existing rules to filter only new recommendations
def merge_tool(aws_data, s3_bucket, it_tool_recommendation_key, existing_it_tool_key):
    try:
        logger.info("START: merge_tool")
        recommendations_df = aws_data.get_dataframe_from_s3(
            s3_bucket, it_tool_recommendation_key
        )
        existing_df = aws_data.get_dataframe_from_s3(s3_bucket, existing_it_tool_key)
        # rename column
        existing_cols = existing_df.columns.values
        existing_cols[0] = "value1"
        existing_df.columns = existing_cols
        # set indexes
        recommendations_df.set_index(["value1", "org_id", "env"], inplace=True)
        existing_df["exists"] = 1
        existing_df.set_index(["value1", "org_id", "env"], inplace=True)
        # join on existing
        recommendations_df = recommendations_df.join(
            existing_df, on=["value1", "org_id", "env"]
        )
        # remove items that are already existing
        recommendations_df = recommendations_df[recommendations_df["exists"].isna()]
        recommendations_df = recommendations_df.drop(columns=["exists"]).reset_index()
        logger.info("END: merge_tool")
        return recommendations_df
    except Exception as e:
        columns = [
            "recommendation_type",
            "value1",
            "value2",
            "value3",
            "env",
            "org_id",
            "policy_id",
            "dev_cnt",
            "org_cnt",
            "alert_cnt",
            "alert_dev_cnt",
            "comment1",
            "comment2",
            "impact_score",
        ]
        logger.error(f"Failed in merge_tool. Error details: {e}")
        return pd.DataFrame(columns=columns)
        # TODO: IT_TOLL might be empty, disable exception temporarily
        # raise Exception(f"Failed in merge_tool. Error details: {e}")


def get_org_id_env(df):
    try:
        df["org_key"] = df["org_key"].fillna("").astype(str)
        df["org_id"] = np.where(
            df["org_key"] != "",
            df["org_key"].apply(
                lambda x: decode_orgkey(orgkey=x).org_id if x.strip() else ""
            ),
            df["org_id"],
        )
        df["env"] = np.where(
            df["org_key"] != "",
            df["org_key"].apply(
                lambda x: decode_orgkey(orgkey=x).env if x.strip() else ""
            ),
            df["env"],
        )
        return df
    except KeyError as e:
        print(
            f"KeyError: {e} - DataFrame does not have the 'org_key' column. Ignorable during transition"
        )
        return df


# Combine all the recommendation types and generate impact score
def gen_recommendations_from_s3(
    s3_bucket,
    hash_events_s3_key,
    hash_pool_s3_key,
    existing_hash_rules_s3_key,
    cert_events_s3_key,
    cert_pool_s3_key,
    existing_cert_rules_s3_key,
    it_tool_recommendation_key,
    existing_it_tool_key,
    output_recommendations_key,
):
    aws_data = aws_s3.AwsS3({}, cross_account_arn=CROSS_ACCOUNT_ARN)
    # combine the hash, cert, and it_tool recommendations into a single dataframe
    recommendation_df = pd.concat(
        [
            merge_hash(
                aws_data,
                s3_bucket,
                hash_events_s3_key,
                hash_pool_s3_key,
                existing_hash_rules_s3_key,
            ),
            merge_cert(
                aws_data,
                s3_bucket,
                cert_events_s3_key,
                cert_pool_s3_key,
                existing_cert_rules_s3_key,
            ),
            merge_tool(
                aws_data, s3_bucket, it_tool_recommendation_key, existing_it_tool_key
            ),
        ],
        ignore_index=True,
    )
    logger.info("START: generate impact score")
    recommendation_df = generate_impact_score(recommendation_df)
    logger.info("END: generate impact score")
    logger.debug(f"recommendation_df: info: {recommendation_df.info()}")

    # re-order the columns into the order that the insert into backend expects
    cols = [
        "recommendation_type",
        "value1",
        "value2",
        "value3",
        "env",
        "org_id",
        "policy_id",
        "dev_cnt",
        "org_cnt",
        "alert_cnt",
        "alert_dev_cnt",
        "comment1",
        "comment2",
        "impact_score",
    ]
    recommendation_df = recommendation_df[cols]
    recommendation_df["org_id"] = recommendation_df["org_id"].astype(int)
    recommendation_df["org_cnt"] = recommendation_df["org_cnt"].astype(int)
    recommendation_df["alert_cnt"] = recommendation_df["alert_cnt"].astype(int)
    # save the dataframe as csv file in s3
    aws_data.dataframe_to_s3(recommendation_df, s3_bucket, output_recommendations_key)


def send_request_to_api(env, org_id, recommendation_json):
    try:
        logger.info(f"env = {env} , org_id = {org_id}")
        encoded_env_org = encode_orgkey(env=env, org_id=org_id)
        logger.info(f"encoded_env_org = {encoded_env_org}")
        logger.info(f"Base URL : {INTERNAL_INGRESS_URL}")
        url = f"https://{INTERNAL_INGRESS_URL}/recommendation-service/v1/orgs/{encoded_env_org}/recommendation"
        logger.info(f"Calling API : {url}")
        headers = {
            "Content-type": "application/json",
            "X-Auth-Token": SERVICE_TOKEN,
        }
        logger.info(
            f"Calling API to populate recommendations with JSON body : {recommendation_json}"
        )
        response = requests.post(url, data=recommendation_json, headers=headers)
        logger.info("Successfully inserted recommendations")
        logger.debug(f"Response: {response}")
    except Exception as e:
        traceback.print_exc()
        logger.error(f"Error in inserting a recommendation to API. Details: {e}")


def insert_recommendations_into_backend_postgres(
    s3_bucket,
    output_recommendations_key,
    hash_pool_key,
    cert_pool_key,
    it_tool_pool_key,
):
    org_cnt_normalizer_types = ["it_tools", "hash", "cert"]
    org_cnt_normalizer = org_num_normalizer(
        org_cnt_normalizer_types,
        prevalence_thresh=PREVALENCE_THRESH,
        s3_bucket=s3_bucket,
        hash_pool_key=hash_pool_key,
        cert_pool_key=cert_pool_key,
        it_tool_pool_key=it_tool_pool_key,
    )
    # load existing recommendations
    aws_data = aws_s3.AwsS3({}, cross_account_arn=CROSS_ACCOUNT_ARN)
    logger.info(
        f"Downloading from S3 output_recommendations_key {output_recommendations_key}"
    )
    recommendation_df = aws_data.get_dataframe_from_s3(
        s3_bucket, output_recommendations_key, escapechar="\\"
    )
    for _, recommendation in recommendation_df.iterrows():
        (
            recommendation_type,
            value1,
            value2,
            value3,
            env,
            org_id,
            policy_id,
            dev_cnt,
            org_cnt,
            alert_cnt,
            alert_dev_cnt,
            comment1,
            comment2,
            impact_score,
        ) = recommendation
        if recommendation_type == "perm_rule":
            rule_type = "sensor_policy"
            perm_json = create_perm_recomm_json(
                rule_type,
                env,
                org_id,
                policy_id,
                value1,
                value2,
                value3,
                org_cnt_normalizer.get_normalized_org_cnt(org_cnt, recommendation_type),
                dev_cnt,
                alert_cnt,
            )
            logger.debug(f"env: {env}, org_id: {org_id}, perm_json: {perm_json}")
            send_request_to_api(env, org_id, perm_json)
        else:
            rule_type = "reputation_override"
            if recommendation_type == "cert":
                r_type = "CERT"
            elif recommendation_type == "hash":
                r_type = "SHA256"
            elif recommendation_type == "it_tools":
                r_type = "IT_TOOL"
            else:
                raise TypeError(
                    f"Error: recommendation_type error: {recommendation_type}"
                )
            reputation_json = create_reputation_recomm_json(
                rule_type,
                env,
                org_id,
                value1,
                value2,
                org_cnt_normalizer.get_normalized_org_cnt(org_cnt, recommendation_type),
                dev_cnt,
                alert_cnt,
                r_type,
                comment2,
                impact_score,
            )
            logger.debug(
                f"env: {env}, org_id: {org_id}, reputation_json: {reputation_json}"
            )
            send_request_to_api(env, int(org_id), reputation_json)


def gen_wf_rec_stats(s3_bucket, output_recommendations_key, wf_server, wf_source):
    # load existing recommendations
    logger.info(f"Sending metrics to {wf_server} , source = {wf_source}")
    aws_data = aws_s3.AwsS3({}, cross_account_arn=CROSS_ACCOUNT_ARN)
    recommendation_df = aws_data.get_dataframe_from_s3(
        s3_bucket, output_recommendations_key
    )
    wf_metrics = []
    now_timestamp = round(time.time() * 1000)
    # get the counts for each org
    stats = recommendation_df.groupby(["org_id", "env", "recommendation_type"]).size()
    for index, value in stats.items():
        wf_metrics.append(
            metric_to_line_data(
                name="policy-suggestion.recommendations",
                value=value,
                timestamp=now_timestamp,
                source=wf_source,
                tags={"org_id": f"{index[0]}", "env": index[1], "type": index[2]},
                default_source=wf_source,
            )
        )

    client_factory = WavefrontClientFactory()
    client_factory.add_client(
        url=wf_server,
        max_queue_size=50000,
        batch_size=10000,
        flush_interval_seconds=1,
    )
    wf_client = client_factory.get_client()
    logger.debug(f"metric data : {wf_metrics}")
    wf_client.send_metric_now(wf_metrics)


def empty_hash_sugg_gen(generate_date):
    hash_pool_s3_bucket = RULES_POOL_BUCKET
    empty_pool_file_path = f"{generate_date}/rules_pool/proc_sha256hash/hash_list_with_common_path_and_file_name.csv"
    hash_pool_s3_location = f"s3://{hash_pool_s3_bucket}/{empty_pool_file_path}"
    columns = [
        "SHA256_HASH",
        "device_count,org_count",
        "device_count_percent",
        "region",
        "proc_path",
        "file_name",
    ]
    df = pd.DataFrame(columns=columns)
    df.to_csv(hash_pool_s3_location, index=False)
    return None


def main():
    """
    Main method to generate suggestions for multiple orgs

    :return: A list of suggestion files
    """
    generate_date = "2020-09-06"
    stage = "bootstrap"
    start_time = time.time()
    logger.debug(
        "################ CLEARING and RECREATING DATABASE TABLES.... ################"
    )
    create_rds_tables()
    logger.debug("CLEARING and RECREATING DATABASE TABLES DONE")
    logger.debug("################ GENERATING hash recommendations ###############")
    gen_hash_sug(generate_date=generate_date, stage=stage)
    logger.debug(
        "################ GENERATING hash recommendations DONE ###############"
    )
    logger.debug("################ GENERATING cert recommendations ###############")
    gen_cert_sug(generate_date=generate_date, stage=stage)
    logger.debug(
        "################ GENERATING cert recommendations DONE ###############"
    )
    logger.debug(
        "################ GENERATING perm_rule recommendations ###############"
    )
    gen_proc_rules_sug(
        generate_date=generate_date,
        generation_type="perm_rule",
        stage=stage,
    )
    logger.debug(
        "################ GENERATING perm_rule recommendations DONE ###############"
    )
    logger.debug("################ GENERATING it_tools recommendations ###############")
    gen_proc_rules_sug(
        generate_date=generate_date,
        generation_type="it_tools",
        stage=stage,
    )
    logger.debug(
        "################ GENERATING it_tools recommendations DONE ###############"
    )
    duration = time.time() - start_time
    logger.debug(f"Time used for generating all types of recommendations: {duration}")


if __name__ == "__main__":
    main()
