#!/usr/bin/env python
# coding: utf-8

import os
from datetime import datetime, timedelta

from mycroft.common import gen_signature, validate_date
from mycroft.common.aws_s3 import AwsS3
from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
STAGE = os.getenv("STAGE", "dev")
logger = get_logger(LOGLEVEL)


def gen_monthly_events_with_impact_number(
    dev_type="WINDOWS",
    generate_date=None,
    reason_subject="proc_sha256hash",
    db_name="proc_hash_cert_alert_db",
    tbl_name="proc_hash_cert_daily_alert_dev_count",
    region="us-east-1",
    query_bucket="cdca-policy-suggest-analytics-dev-athena-outputs",
    output_bucket="cdca-dev-us-east-1-intermediate-data",
    output_folder="2020-09-06/events_pool/proc_sha256hash/",
    alert_cnt_filter=1,
    dev_cnt_filter=5,
    days=30,
    cross_account_arn=None,
):
    """
    Generates proc_sha256 data with reasons for a given org_id.
    :param dev_type: Device type (OS type)
    :param generate_date: Date the recommendations being created
    :param reason_subject: Can be proc_sha256, proc_path, it_tools or cert
    :param db_name: Database name: proc_hash_cert_alert_db
    :param tbl_name: Table name: proc_hash_cert_daily_alert_dev_count
    :param region: AWS region such as "us-east-1",
    :param query_bucket: Query result bucket,
    :param output_bucket: Query output bucket,
    :param output_folder: Bucket to store csv file
    :param alert_cnt_filter: alert count filter
    :param dev_cnt_filter: device count filter
    :param days: Number of days to compute impact number
    :param cross_account_arn: Cross account role ARN
    :return: The s3 path of the generated csv file, if failed, return None
    """
    try:
        assert validate_date(generate_date) is None
        start_date = datetime.strptime(generate_date, "%Y-%m-%d") - timedelta(days=days)
        logger.info(
            f"START: Generating monthly events with impact number for {reason_subject} "
            f"between {start_date} and {generate_date}"
        )
        logger.info(f"Data source: [{db_name}.{tbl_name}]")
        params = {
            "region": region,
            "database": db_name,
            "bucket": query_bucket,
            "outputBucket": output_bucket,
            "outputFolder": output_folder,
            "outputLocation": "s3://" + output_bucket + "/" + output_folder + "/",
        }
        logger.debug(f"params: {params}")
        aws_data = AwsS3(params, cross_account_arn=cross_account_arn)
        query = None
        if reason_subject == "cert":
            query = get_cert_query(
                db_name,
                tbl_name,
                start_date,
                generate_date,
                dev_type,
                dev_cnt_filter,
                alert_cnt_filter,
            )
            logger.debug(f"cert query: {query}")
        elif reason_subject == "proc_sha256hash":
            query = get_proc_sha256hash_query(
                db_name,
                tbl_name,
                start_date,
                generate_date,
                dev_type,
                dev_cnt_filter,
                alert_cnt_filter,
            )
            logger.debug(f"proc_sha256hash query: {query}")
        elif reason_subject == "it_tools":
            query = get_it_tool_query(
                db_name,
                tbl_name,
                start_date,
                generate_date,
                dev_type,
                dev_cnt_filter,
                alert_cnt_filter,
            )
            logger.debug(f"query: {query}")
        s3_file_path, s3_output_bucket, s3_output_key = aws_data.get_res_on_s3(
            query, "all_events.csv", force_query=True
        )
        logger.info(
            f"END: Generating [{days}] events with impact number for {reason_subject} "
            f"Wrote result to {s3_file_path} on {generate_date}"
        )
        return s3_file_path, s3_output_bucket, s3_output_key
    except ValueError as e:
        logger.error(f"Value error. Details: {str(e)}")
        return None, None, None
    except Exception as e:
        logger.error(f"Error in gen_events_with_impact_number. Details: {str(e)}")
        return None, None, None


# This method is never executed.
def gen_weekly_events_with_impact_number(
    dev_type="WINDOWS",
    start_date="2020-08-06",
    generate_start_date="2020-08-18",
    generate_end_date="2020-08-24",
    reason_subject="proc_path",
    db_name="proc_hash_cert_alert_db",
    tbl_name="proc_hash_cert_daily_alert_dev_count",
    region="us-east-1",
    query_bucket="aws-athena-query-results-prod-us-east-1-shannon",
    output_bucket="cdca-prod-us-east-1-intermediate-data",
):
    """
    THIS METHOD IS NOT USED BY EXISTING PIPELINE
    """
    try:
        assert validate_date(start_date) is None
        assert validate_date(generate_start_date) is None
        assert validate_date(generate_end_date) is None
        logger.debug(f"reason_subject {reason_subject}")
        logger.debug(f"Data source: [{db_name}.{tbl_name}]")
        logger.debug(
            f"start_date: {generate_start_date}, end_date: {generate_end_date}"
        )
        if reason_subject == "proc_path":
            output_folder = generate_start_date + "/" + reason_subject + "/events_data"
            params = {
                "region": region,
                "database": db_name,
                "bucket": query_bucket,
                "outputBucket": output_bucket,
                "outputFolder": output_folder,
                "outputLocation": "s3://" + output_bucket + "/" + output_folder + "/",
            }
            logger.debug(f"params: {params}")
            aws_data = AwsS3(params)
            query = f""" SELECT t.env AS env,
                         t.org_id AS org_id,
                         t.policy_id AS policy_id,
                         t.proc_path AS proc_path,
                         t.dev_cnt AS dev_cnt,
                         t.alert_cnt AS alert_cnt
                        FROM
                            (SELECT env,
                             org_id,
                             policy_id,
                             lower(proc_path) AS proc_path,
                            CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                           sum(alert_cnt) AS alert_cnt
                        FROM {db_name}.{tbl_name}
                        WHERE event_date BETWEEN '{start_date}' AND '{generate_end_date}'
                          AND dev_type = '{dev_type}' AND alert_cnt > 0 AND CONCAT(lower(proc_path), '-', env, '-', cast(org_id AS VARCHAR)) IN
                          (SELECT DISTINCT CONCAT(lower(proc_path), '-', env, '-', cast(org_id AS VARCHAR))
                          FROM {db_name}.{tbl_name}
                          WHERE event_date between '{generate_start_date}' and '{generate_end_date}' AND dev_type = '{dev_type}')
                        GROUP BY 1,2,3,4
                        ORDER BY 1 DESC) AS t
                        WHERE t.dev_cnt >= 5"""
            logger.debug(f"query: {query}")
            s3_file_path, s3_output_bucket, s3_output_key = aws_data.get_res_on_s3(
                query, "all_events.csv", force_query=True
            )
            return params["outputBucket"], params["outputFolder"] + "/all_events.csv"
    except ValueError as e:
        logger.error(f"Value error. Details: {str(e)}")
        return None
    except Exception as e:
        logger.error(f"Error in gen_events_with_impact_number. Details: {str(e)}")
        return None


def gen_daily_events_with_impact_number(
    dev_type="WINDOWS",
    generate_date="2020-09-06",
    reason_subject="proc_path",
    db_name="proc_hash_cert_alert_db",
    tbl_name="proc_hash_cert_daily_alert_dev_count",
    region="us-east-1",
    query_bucket="aws-athena-query-results-prod-us-east-1-shannon",
    output_bucket="cdca-prod-us-east-1-intermediate-data",
    output_folder="2020-09-06/events_pool/proc_path",
    alert_cnt_filter=1,
    dev_cnt_filter=5,
):
    """
    THIS METHOD IS NOT USED BY EXISTING PIPELINE
    Generates proc_sha256 data with reasons for a given org_id.
    :param dev_type: Device type (OS type)
    :param generate_date: Date the recommendations being created
    :param reason_subject: Can be proc_sha256, proc_path, it_tools or cert
    :param db_name: Database name: proc_hash_cert_alert_db
    :param tbl_name: Table name: proc_hash_cert_daily_alert_dev_count
    :param region: AWS region such as "us-east-1",
    :param query_bucket: Query result bucket,
    :param output_bucket: Query output bucket,
    :param output_folder: Bucket to store csv file
    :param alert_cnt_filter: alert count filter
    :param dev_cnt_filter: device count filter
    :return: The s3 path of the generated csv file, if failed, return None
    """
    try:
        assert validate_date(generate_date) is None
        start_date = datetime.strftime(
            datetime.strptime(generate_date, "%Y-%m-%d") - timedelta(days=29),
            "%Y-%m-%d",
        )
        oldest_date = datetime.strftime(
            datetime.strptime(generate_date, "%Y-%m-%d") - timedelta(days=30),
            "%Y-%m-%d",
        )
        logger.debug(f"reason_subject {reason_subject}")
        logger.debug(f"Data source: [{db_name}.{tbl_name}]")
        logger.debug(
            f"start_date: {start_date}, oldest_date: {oldest_date}, end_date: {generate_date}"
        )
        params = {
            "region": region,
            "database": db_name,
            "bucket": query_bucket,
            "outputBucket": output_bucket,
            "outputFolder": output_folder,
            "outputLocation": "s3://" + output_bucket + "/" + output_folder + "/",
        }
        logger.debug(f"params: {params}")
        aws_data = AwsS3(params)
        if reason_subject == "cert":
            query = f""" SELECT t.env AS env,
                                t.org_id AS org_id,
                                t.{reason_subject}_signedby AS {reason_subject}_signedby,
                                t.{reason_subject}_authority AS {reason_subject}_authority,
                                t.dev_cnt AS dev_cnt,
                                t.alert_cnt AS alert_cnt
                        FROM (SELECT env,
                             org_id,
                             {reason_subject}_signedby,
                             {reason_subject}_authority,
                             CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                             sum(alert_cnt) AS alert_cnt
                    FROM {db_name}.{tbl_name}
                    WHERE event_date
                        BETWEEN '{start_date}'
                            AND '{generate_date}'
                            AND CONCAT(cert_signedby, '-', cert_authority) IN
                        (SELECT DISTINCT CONCAT(cert_signedby,
                            '-', cert_authority)
                        FROM proc_hash_cert_daily_alert_dev_count
                        WHERE event_date IN ('{generate_date}', '{oldest_date}'))
                    GROUP BY  1,2,3,4) AS t
                    WHERE t.dev_cnt >= {dev_cnt_filter} AND t.alert_cnt >= {alert_cnt_filter}"""
            logger.debug(f"query: {query}")
            s3_file_path = aws_data.get_res_on_s3(
                gen_signature(query, "policy-automation"),
                "all_events.csv",
                force_query=True,
            )
            return s3_file_path
        elif reason_subject == "proc_path":
            query = f"""SELECT  t.env AS env,
                                t.org_id AS org_id,
                                t.policy_id AS policy_id,
                                t.proc_path AS proc_path,
                                t.dev_cnt AS dev_cnt,
                                t.alert_cnt AS alert_cnt
                        FROM (SELECT env,
                                 org_id,
                                 policy_id,
                                 lower(proc_path) AS proc_path,
                                CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                               sum(alert_cnt) AS alert_cnt
                            FROM {db_name}.{tbl_name}
                            WHERE event_date BETWEEN '{start_date}' AND '{generate_date}'
                              AND dev_type = '{dev_type}' AND CONCAT(lower(proc_path), '-', env, '-', cast(org_id AS VARCHAR), '-', cast(policy_id AS VARCHAR)) IN
                              (SELECT DISTINCT CONCAT(lower(proc_path), '-', env, '-', cast(org_id AS VARCHAR), '-', cast(policy_id AS VARCHAR))
                              FROM {db_name}.{tbl_name}
                              WHERE event_date IN ('{generate_date}', '{oldest_date}') AND dev_type = '{dev_type}')
                            GROUP BY 1,2,3,4
                            ORDER BY 1,2,3) AS t
                        WHERE t.dev_cnt >= {dev_cnt_filter} AND t.alert_cnt >= {alert_cnt_filter} """
            logger.debug(f"query: {query}")
            s3_file_path = aws_data.get_res_on_s3(
                gen_signature(query, "policy-automation"),
                "all_events.csv",
                force_query=True,
            )
            return s3_file_path
        elif reason_subject == "it_tools":
            query = f""" SELECT t.env AS env,
                                t.org_id AS org_id,
                                t.proc_path AS proc_path,
                                t.dev_cnt AS dev_cnt,
                                t.alert_cnt AS alert_cnt
                        FROM (SELECT env,
                                 org_id,
                                 lower(proc_path) AS proc_path,
                                CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                               sum(alert_cnt) AS alert_cnt
                            FROM {db_name}.{tbl_name}
                            WHERE event_date BETWEEN '{start_date}' AND '{generate_date}'
                              AND dev_type = '{dev_type}' AND CONCAT(lower(proc_path), '-', env, '-', cast(org_id AS VARCHAR)) IN
                              (SELECT DISTINCT CONCAT(lower(proc_path), '-', env, '-', cast(org_id AS VARCHAR))
                              FROM {db_name}.{tbl_name}
                              WHERE event_date IN ('{generate_date}', '{oldest_date}') AND dev_type = '{dev_type}')
                            GROUP BY 1,2,3
                            ORDER BY 1,2) AS t
                        WHERE t.dev_cnt >= {dev_cnt_filter} AND t.alert_cnt >= {alert_cnt_filter}"""
            logger.debug(f"query: {query}")
            s3_file_path = aws_data.get_res_on_s3(
                gen_signature(query, "policy-automation"),
                "all_events.csv",
                force_query=True,
            )
            return s3_file_path
        elif reason_subject == "proc_sha256hash":
            query = f""" SELECT t.env AS env,
                                t.org_id AS org_id,
                                t.{reason_subject} AS {reason_subject},
                                t.dev_cnt AS dev_cnt,
                                t.alert_cnt AS alert_cnt
                         FROM (SELECT env,
                                    org_id,
                                    {reason_subject},
                                    CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                                    sum(alert_cnt) AS alert_cnt
                               FROM {db_name}.{tbl_name}
                               WHERE event_date BETWEEN '{start_date}' AND '{generate_date}'
                                    AND CONCAT({reason_subject}, '-', env, '-', cast(org_id AS VARCHAR)) IN
                                    (SELECT DISTINCT CONCAT({reason_subject}, '-', env, '-', cast(org_id AS VARCHAR))
                                    FROM {db_name}.{tbl_name}
                                    WHERE event_date IN ('{generate_date}', '{oldest_date}'))
                            GROUP BY 1,2,3) AS t
                        WHERE t.dev_cnt >= {dev_cnt_filter} AND t.alert_cnt >= {alert_cnt_filter} """
            logger.debug(f"query: {query}")
            s3_file_path, s3_output_bucket, s3_output_key = aws_data.get_res_on_s3(
                query, "all_events.csv", force_query=True
            )
            return s3_file_path, s3_output_bucket, s3_output_key
        else:
            raise ValueError(
                "reason_subject value error: {reason_subject}. It can only be (proc_path, cert, proc_sha256hash)."
            )
    except ValueError as e:
        logger.error(f"Value error. Details: {str(e)}")
        return None
    except Exception as e:
        logger.error(f"Error in gen_events_with_impact_number. Details: {str(e)}")
        return None


def get_cert_query(
    db_name,
    tbl_name,
    start_date,
    generate_date,
    dev_type,
    dev_cnt_filter,
    alert_cnt_filter,
):
    logger.info(
        f"Prepare SHA256HASH query for UDS stream table in [{STAGE}] region on [{generate_date}]"
    )
    return f"""SELECT t.env AS env,
                             t.org_id AS org_id,
                             t.cert_signedby AS cert_signedby,
                             t.cert_authority AS cert_authority,
                             t.dev_cnt AS dev_cnt,
                             t.alert_cnt AS alert_cnt,
                             t.org_key as org_key
                    FROM
                        (SELECT env,
                            org_id,
                            cert_signedby,
                            cert_authority,
                            CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                            sum(alert_cnt) AS alert_cnt,
                            org_key AS org_key
                        FROM {db_name}.{tbl_name}
                        WHERE event_date
                            BETWEEN '{start_date}' AND '{generate_date}'
                            AND dev_type = '{dev_type}'
                            AND regexp_like(org_key, '^[A-Z0-9]+$')
                        GROUP BY 1,2,3,4,7) As t
                        WHERE t.dev_cnt >= {dev_cnt_filter}
                        AND t.alert_cnt >= {alert_cnt_filter}"""


def get_proc_sha256hash_query(
    db_name,
    tbl_name,
    start_date,
    generate_date,
    dev_type,
    dev_cnt_filter,
    alert_cnt_filter,
):
    logger.info(
        f"Prepare SHA256HASH query for UDS stream table in [{STAGE}] region on [{generate_date}]"
    )
    return f"""SELECT t.env AS env,
                             t.org_id AS org_id,
                             t.proc_sha256hash AS proc_sha256hash,
                             t.dev_cnt AS dev_cnt,
                             t.alert_cnt AS alert_cnt,
                             t.org_key as org_key
                    FROM
                        (SELECT env,
                             org_id,
                             proc_sha256hash,
                            CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                           sum(alert_cnt) AS alert_cnt,
                           org_key AS org_key
                        FROM {db_name}.{tbl_name}
                        WHERE event_date BETWEEN '{start_date}' AND '{generate_date}'
                              AND dev_type = '{dev_type}'
                              AND regexp_like(org_key, '^[A-Z0-9]+$')
                        GROUP BY 1,2,3,6) AS t
                        WHERE t.dev_cnt >= {dev_cnt_filter}
                        AND t.alert_cnt >= {alert_cnt_filter}"""


def get_it_tool_query(
    db_name,
    tbl_name,
    start_date,
    generate_date,
    dev_type,
    dev_cnt_filter,
    alert_cnt_filter,
):
    logger.info(
        f"Prepare IT_TOOL query for UDS stream table in [{STAGE}] region on [{generate_date}]"
    )
    return f"""SELECT t.env AS env,
                     t.org_id AS org_id,
                     t.proc_path AS proc_path,
                     t.dev_cnt AS dev_cnt,
                     t.alert_cnt AS alert_cnt,
                     t.org_key as org_key
            FROM
                (SELECT env,
                     org_id,
                     lower(proc_path) AS proc_path,
                     CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS dev_cnt,
                     sum(alert_cnt) AS alert_cnt,
                     org_key AS org_key
                FROM {db_name}.{tbl_name}
                WHERE event_date
                    BETWEEN '{start_date}'
                        AND '{generate_date}'
                        AND dev_type = '{dev_type}'
                        AND alert_cnt >= {alert_cnt_filter}
                        AND (lower(proc_path) LIKE 'c:\\program files (x86)\\%'
                            or lower(proc_path) LIKE 'c:\\program files\\%'
                            or lower(proc_path) LIKE 'c:\\programdata\\%'
                            or lower(proc_path) LIKE 'c:\\windows\\ccm%'
                            or lower(proc_path) LIKE 'c:\\windows\\installer\\%'
                            or lower(proc_path) LIKE 'c:\\windows\\veeam%')
                        AND regexp_like(org_key, '^[A-Z0-9]+$')
                GROUP BY  1,2,3,6
                ORDER BY  1,2 DESC) AS t
            WHERE t.dev_cnt >= {dev_cnt_filter}"""
