import csv
import os
from typing import Dict, List

from smart_open import open as smart_open

from mycroft.common.pathtree.path_tree import PathTree
from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


def write(group_rules: list, grouped_rules_path: str, csv_columns: List[str]):
    with smart_open(grouped_rules_path, "w") as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=csv_columns)
        writer.writeheader()
        for entry in group_rules:
            writer.writerow(entry)


def read_suggestion_pool(rule_sug_pool_file: str, pool_type: str) -> Dict:
    logger.debug(f"rule_sug_pool_file: {rule_sug_pool_file}")

    # `results` can be different based on the pool type
    results = dict()  # type: ignore
    permission = True if pool_type == "permission" else False
    if permission:
        with smart_open(rule_sug_pool_file, encoding="utf-8") as f:
            csv_reader = csv.DictReader(f, delimiter=",")
            # skip header if needed
            next(csv_reader, None)
            for line in csv_reader:
                # Group rule path by operation and action tuple
                operation_action = line["operation"] + "," + line["action"]
                if operation_action not in results:
                    results[operation_action] = []
                    results[operation_action].append(
                        (line["application_value"], line["org_cnt"])
                    )
                else:
                    results[operation_action].append(
                        (line["application_value"], line["org_cnt"])
                    )
    else:
        with smart_open(rule_sug_pool_file, encoding="utf-8") as f:
            csv_reader = csv.DictReader(f, delimiter=",")
            # skip header if needed
            next(csv_reader, None)
            for line in csv_reader:
                logger.debug("line_fields %s", line)
                # Skip malformed lines
                if len(line) < 2:
                    continue
                # Build a tree from the rule path
                rule_path = line["application_value"].strip('\n,"')
                # Initialize the group ccount and the prevalence sum
                # Initialize the aggregater dictionary
                # Values are group rule path count, total prevalence, list of rule path in the group,
                # initial prevalence of the rule path
                results[rule_path] = [
                    0,
                    0,
                    [],
                    False,
                    int(line["org_cnt"].strip('\n,"')),
                ]
    return results


def group_perm_rules(
    operation_action_dict, bad_rule_paths, remove_start_dbl_str, grp_prev_thresh,
):
    group_rules = []
    for operation_action, vals in operation_action_dict.items():
        # build a tree with the list
        group_path_tree = PathTree("root")
        rule_path_group_info = dict()
        for rule_entry in vals:
            rl_path = rule_entry[0].strip('\n,"')
            group_path_tree.build_tree(rl_path)
            # Initialize the group ccount and the prevalence sum
            rule_path_group_info[rl_path] = [0, 0, [], False]

        # Do matching now that tree is built
        for rule_entry in vals:
            rl_path = rule_entry[0].strip('\n,"')
            (
                gr_pm_list,
                num_visits,
                num_path_tokens,
            ) = group_path_tree.find_path_in_tree(rl_path)

            for grouping_path in gr_pm_list:
                logger.debug("grouping_path %s", grouping_path)
                first_rparent_token = grouping_path.split("\\")[0]

                if grouping_path in bad_rule_paths:
                    continue
                if grouping_path in rule_path_group_info:
                    rule_path_group_info[grouping_path][0] += 1
                    prevalence_val = int(rule_entry[1].strip('\n,"'))
                    rule_path_group_info[grouping_path][1] += prevalence_val
                    rule_path_group_info[grouping_path][2].append(rl_path)
                # If a path tree match is found for a rule path and if that match
                # is different from the rule path, the rule path is
                # in a wider rule path group.
                # So it cannot represent the group, so mark it to exclude from selection
                if not rule_path_group_info[rl_path][3] and grouping_path != rl_path:
                    rule_path_group_info[rl_path][3] = True
                    if remove_start_dbl_str and first_rparent_token == "**":
                        rule_path_group_info[rl_path][3] = False

        for rpk in rule_path_group_info:
            first_rparent_token = rpk.split("\\")[0]
            if not rule_path_group_info[rpk][3]:
                if remove_start_dbl_str and first_rparent_token == "**":
                    continue
                grp_prev = rule_path_group_info[rpk][1]
                if grp_prev > grp_prev_thresh:
                    group_line = {
                        "rpk": rpk,
                        "operation_action": operation_action,
                        "grp_prev": str(grp_prev),
                    }
                    group_rules.append(group_line)

    return group_rules


def group_rule_it_tool(
    group_path_tree,
    rule_path_group_info,
    bad_rule_paths,
    remove_start_dbl_str,
    grp_prev_thresh,
):
    group_rules = []

    # Now that the path tree is built, group the rule pathes summing up their prevaence
    for rl_path in rule_path_group_info:
        # Get the matches for this rule path
        gr_pm_list, num_visits, num_path_tokens = group_path_tree.find_path_in_tree(
            rl_path
        )
        for grouping_path in gr_pm_list:
            logger.debug("grouping_path %s", grouping_path)
            # TODO fix to make Mac/Linux aware
            first_rparent_token = grouping_path.split("\\")[0]

            if grouping_path in bad_rule_paths:
                continue
            # If not it is matched and set to 0
            if grouping_path in rule_path_group_info:
                rule_path_group_info[grouping_path][0] += 1
                prevalence_val = int(rule_path_group_info[rl_path][4])
                rule_path_group_info[grouping_path][1] += prevalence_val
                rule_path_group_info[grouping_path][2].append(rl_path)
            # If a path tree match is found for a rule path and if that match
            # is different from the rule path, the rule path is
            # in a wider rule path group.
            # So it cannot represent the group, so mark it to exclude from selection
            if (
                grouping_path in rule_path_group_info
                and not rule_path_group_info[grouping_path][3]
                and grouping_path != rl_path
            ):
                rule_path_group_info[grouping_path][3] = True
                if remove_start_dbl_str and first_rparent_token == "**":
                    rule_path_group_info[grouping_path][3] = False

    for rpk in rule_path_group_info:
        # TODO fix to make Mac/Linux aware
        first_rparent_token = rpk.split("\\")[0]
        if not rule_path_group_info[rpk][3]:
            if remove_start_dbl_str and first_rparent_token == "**":
                continue
            grp_prev = rule_path_group_info[rpk][1]
            if grp_prev > grp_prev_thresh:
                group_line = {
                    "rpk": rpk,
                    "grp_prev": str(grp_prev),
                }
                group_rules.append(group_line)
    return group_rules


def group_it_tools_path(
    rule_sug_pool_file,
    grouped_rules_path,
    grp_prev_thresh=1,
    remove_start_dbl_str=False,
    bad_rule_paths=None,
):
    rule_path_group_info = read_suggestion_pool(rule_sug_pool_file, "it_tool")
    group_path_tree = PathTree("root")
    for rl_path in rule_path_group_info:
        logger.debug(rl_path)
        group_path_tree.build_tree(rl_path)
    # Now that the path tree is built, group the rule pathes summing up their prevaence
    grouped_rules = group_rule_it_tool(
        group_path_tree,
        rule_path_group_info,
        bad_rule_paths,
        remove_start_dbl_str,
        grp_prev_thresh,
    )

    # def write(group_rules: list, grouped_rules_path: str, csv_columns: List[str]):
    csv_columns = ["rpk", "grp_prev"]
    write(grouped_rules, grouped_rules_path, csv_columns)

    return grouped_rules_path


def group_perm_rules_path(
    rule_sug_pool_file,
    grouped_rules_path,
    grp_prev_thresh: int = 1,
    remove_start_dbl_str: bool = False,
    bad_rule_paths: list = None,
):
    logger.debug("rule_sug_pool_file: %s", rule_sug_pool_file)
    rule_path_group_info = read_suggestion_pool(rule_sug_pool_file, "permission")
    grouped_rules = group_perm_rules(
        rule_path_group_info, bad_rule_paths, remove_start_dbl_str, grp_prev_thresh,
    )
    # For each operation action tuple, generate a representative group
    # with its aggregated prevalence values
    try:
        csv_columns = ["rpk", "operation_action", "grp_prev"]
        write(grouped_rules, grouped_rules_path, csv_columns)
    except IOError:
        logger.exception("I/O error trying to write csv with grouped paths")
    return grouped_rules_path
