#!/usr/bin/env python
# coding: utf-8

import csv
import os
from typing import List, Optional

from smart_open import open as smart_open
from wcmatch import pathlib

from mycroft.common.aws_s3 import assume_role_session, get_s3_client, get_s3_resource
from mycroft.common.logging import get_logger
from mycroft.generate_suggestions.write_suggestions import write_rules_to_csv
from psc_orgkey import decode_orgkey

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


TEST_PATH_TREE_HERE = False


def normalize_path(path: str, operating_system: Optional[str]) -> pathlib.PurePath:
    if operating_system == "Windows":
        return pathlib.PureWindowsPath(path.lower())
    else:
        return pathlib.PurePath(path)


class TreeNode:
    def __init__(self, p_token):
        self.cnt = 1
        self.token = p_token
        self.child = dict()


class PathTree:
    def __init__(self, p_token):
        self.rule_of = dict()
        self.tree_node = TreeNode(p_token)

    def build_tree(self, proc_path):
        # Create children
        proc_path_tokens = proc_path.split("\\")
        #         num_tokens = len(pro_path_tokens)
        num_tokens = proc_path_tokens.__len__()
        # Clean tokens of \"mark
        if '"' in proc_path_tokens[0]:
            proc_path_tokens[0] = proc_path_tokens[0][1:]
        #         if '\\n"' in proc_path_tokens[num_tokens -1]:
        if "\n" in proc_path_tokens[num_tokens - 1]:
            proc_path_tokens[num_tokens - 1] = proc_path_tokens[num_tokens - 1][:-1]
        if '"' in proc_path_tokens[num_tokens - 1]:
            proc_path_tokens[num_tokens - 1] = proc_path_tokens[num_tokens - 1][:-1]
        # Add the proc_path to the tree in reverse order
        cur_node = self.tree_node
        for i in range(0, num_tokens):
            token = proc_path_tokens[num_tokens - 1 - i]
            if cur_node.child.get(token) is not None:
                cur_node.child.get(token).cnt += 1
            else:
                new_node = TreeNode(token)
                cur_node.child[token] = new_node
            cur_node = cur_node.child[token]

    # stream lines from an S3 object
    def add_rules_to_tree(self, s3_path, cross_account_arn):
        logger.debug("s3_path: %s", s3_path)
        s3_client = get_s3_client(cross_account_arn)
        logger.debug("[add_rules_to_tree] using s3_client to open")
        for line in smart_open(
            s3_path, transport_params={"client": s3_client}, encoding="utf-8"
        ):
            logger.debug(f"read line {line}")
            line_fields = line.split(",")
            path_item = line_fields[0]
            if line_fields[0] and '"' == line_fields[0][0]:
                path_item = line_fields[0][1:-1]
            # Store the rule information for each path
            if len(line_fields) > 1:
                rule_val_str = line_fields[1]
                if len(line_fields) > 2:
                    rule_val_str = line_fields[1] + "," + line_fields[2]
                elif len(line_fields) > 3:
                    rule_val_str = (
                        line_fields[1]
                        + ","
                        + line_fields[2]
                        + ","
                        + line_fields[3][:-1]
                    )
                if self.rule_of.get(path_item) is None:
                    self.rule_of[path_item] = []
                if rule_val_str not in self.rule_of[path_item]:
                    self.rule_of[path_item].append(rule_val_str)
                logger.debug(f"rule_val_str: {rule_val_str}")
            logger.debug(f"line field[0]: {line_fields[0]}")
            self.build_tree(line_fields[0])

    # flake8: noqa: C901
    def find_path_in_tree(self, proc_path):
        # TODO refactor: make linux/macOS compatible
        proc_path_tokens = proc_path.split("\\")
        num_tokens = proc_path_tokens.__len__()
        if '"' in proc_path_tokens[0]:
            proc_path_tokens[0] = proc_path_tokens[0][1:]
        if '\\n"' in proc_path_tokens[num_tokens - 1]:
            proc_path_tokens[num_tokens - 1] = proc_path_tokens[num_tokens - 1][:-1]
        if '"' in proc_path_tokens[num_tokens - 1]:
            proc_path_tokens[num_tokens - 1] = proc_path_tokens[num_tokens - 1][:-1]
        if len(proc_path_tokens[num_tokens - 1]) > 0:
            if (
                '"'
                == proc_path_tokens[num_tokens - 1][
                    len(proc_path_tokens[num_tokens - 1]) - 1
                ]
            ):
                proc_path_tokens[num_tokens - 1] = proc_path_tokens[num_tokens - 1][:-1]
        if len(proc_path_tokens[num_tokens - 1]) > 1:
            if (
                '"'
                == proc_path_tokens[num_tokens - 1][
                    len(proc_path_tokens[num_tokens - 1]) - 2
                ]
            ):
                proc_path_tokens[num_tokens - 1] = proc_path_tokens[num_tokens - 1][:-2]
        if len(proc_path_tokens[num_tokens - 1]) > 2:
            if (
                '"'
                == proc_path_tokens[num_tokens - 1][
                    len(proc_path_tokens[num_tokens - 1]) - 3
                ]
            ):
                proc_path_tokens[num_tokens - 1] = proc_path_tokens[num_tokens - 1][:-3]
        cur_node = self.tree_node
        rule_subject = ""
        dbl_str_in_path = False
        node_of_path = dict()
        node_of_path[rule_subject] = cur_node
        path_str_node = (rule_subject, cur_node, dbl_str_in_path)
        path_queue = []
        node_of_path[rule_subject] = cur_node
        path_queue.append(path_str_node)
        next_path_queue = []
        # stats collection
        num_nodes_visted = 0

        for i in range(0, num_tokens):
            token = proc_path_tokens[num_tokens - 1 - i]
            if next_path_queue:
                path_queue = next_path_queue.copy()
            while path_queue:
                cur_node_of_path = path_queue.pop()
                # Remove from the next queue too
                if next_path_queue:
                    next_path_queue.remove(cur_node_of_path)

                (cur_path, cur_node, dbl_str_in_path) = cur_node_of_path
                next_match_found = False
                if dbl_str_in_path and i < num_tokens - 1:
                    next_path_queue.append(cur_node_of_path)
                    next_match_found = True
                fuzzy_token_part = token.split(" ")[0] + "**"
                for rule_path_token in [
                    token,
                    "*",
                    "**",
                    token + "*",
                    "*:",
                    fuzzy_token_part,
                ]:
                    if rule_path_token in cur_node.child:
                        next_match_found = True
                        next_match_path = rule_path_token + "\\" + cur_path
                        if cur_path == "":
                            next_match_path = rule_path_token
                        next_match_node = cur_node.child[rule_path_token]
                        node_of_path[next_match_path] = next_match_node
                        if rule_path_token == "**":
                            dbl_str_in_path = True
                        next_match_path_node = (
                            next_match_path,
                            next_match_node,
                            dbl_str_in_path,
                        )
                        next_path_queue.append(next_match_path_node)
                        num_nodes_visted += 1
                if next_match_found or (
                    not next_match_found and len(cur_node.child) > 0
                ):
                    if cur_path in node_of_path:
                        del node_of_path[cur_path]

        return node_of_path, num_nodes_visted, num_tokens

    # stream lines from an S3 object
    def find_policy_rules_in_tree(self, s3_path, with_reasons=False):
        selected_rules = dict()
        # stats collection
        num_nodes_visted_arr = []
        num_matching_path_arr = []
        num_path_tokens_arr = []
        # Following commented lines sometimes used to get program stats
        #         num_nodes_visted_dict = dict()
        #         num_matching_path_dict = dict()
        cnt = 0
        for line in smart_open(s3_path, encoding="utf-8"):
            cnt += 1
            line_arr = line.split(",")
            line = line_arr[0]
            list_of_path_match, num_visits, num_path_tokens = self.find_path_in_tree(
                line
            )
            if len(list_of_path_match) > 0:
                # For stats
                num_nodes_visted_arr.append(num_visits)
                num_matching_path_arr.append(len(list_of_path_match))
                num_path_tokens_arr.append(num_path_tokens)
                # Following commented lines sometimes used to get program stats
                # num_nodes_visted_dict[line] = num_visits
                # num_matching_path_dict[line] = len(list_of_path_match)
            for key in list_of_path_match:
                path_match = key
                if path_match is not None:
                    if path_match in self.rule_of:
                        if with_reasons:
                            # First reason is dev_cnt and we take the maximum for now
                            # Second reason is alert count and we take the sum for now
                            try:
                                # TODO Use strip instead of the following conditions
                                if '"' in line_arr[1]:
                                    line_arr[1] = line_arr[1][1:-1]
                                if '"' in line_arr[2]:
                                    line_arr[2] = line_arr[2][1:-2]
                            except IndexError:
                                logger.error("Reason value is invalid")
                            if path_match in selected_rules:
                                # First tuple for recommendation and second tuple contains reasons array
                                cur_reasons_arr = selected_rules[path_match][1]
                                try:
                                    if cur_reasons_arr[0] < int(line_arr[1]):
                                        cur_reasons_arr[0] = int(line_arr[1])
                                    cur_reasons_arr[1] += int(line_arr[2])
                                except Exception as e:
                                    logger.debug(
                                        "Invalid reason value parsed. {0}".format(e)
                                    )
                            else:
                                cur_reasons_arr = []
                                try:
                                    cur_reasons_arr.append(int(line_arr[1]))
                                    cur_reasons_arr.append(int(line_arr[2]))
                                except Exception as e:
                                    logger.debug(
                                        "Invalid reason value parsed. {0}".format(e)
                                    )

                            selected_rules[path_match] = (
                                self.rule_of[path_match],
                                cur_reasons_arr,
                            )
                        #                             logger.info(selected_rules)
                        else:
                            selected_rules[path_match] = self.rule_of[path_match]
        return (
            selected_rules,
            num_nodes_visted_arr,
            num_matching_path_arr,
            num_path_tokens_arr,
        )

    def find_multi_proc_path_in_tree(
        self,
        local_output_file,
        local_temp_file,
        rule_path_match,
        suggestion_type,
        events_bucket,
        events_file_key,
        cross_account_arn,
    ):
        s3 = get_s3_resource(cross_account_arn)
        # s3 = boto3.resource("s3")
        s3.Bucket(events_bucket).download_file(events_file_key, local_temp_file)
        cnt = 0
        cached_matched_rules = dict()
        num_nodes_visted_arr = []
        num_matching_path_arr = []
        num_path_tokens_arr = []
        matched_rules = 0
        prev_env_org_policy = None
        cur_env_org_policy = None
        write_header = True
        with open(local_temp_file, "r") as csvfile:
            datareader = csv.DictReader(csvfile, delimiter=",")
            for line in datareader:
                logger.info(f"find multi path in tree: {line}")
                cnt += 1
                try:
                    decoded = decode_orgkey(orgkey=line["org_key"])
                    env = line["env"] if line["env"] else decoded.env
                    if suggestion_type == "perm_rule":
                        policy_id = int(line["policy_id"])
                    else:
                        policy_id = ""
                    proc_path = line["proc_path"]
                    dev_cnt = int(line["dev_cnt"])
                    alert_cnt = int(line["alert_cnt"])
                    org_id = (
                        int(line["org_id"]) if line["org_id"] else int(decoded.org_id)
                    )
                    cur_env_org_policy = (env, org_id, policy_id)
                    logger.info(
                        f"parsing the event: proc_path: {proc_path}, env: {env}, org_id: {org_id}, policy_id: {policy_id}, dev_cnt: {dev_cnt}, alert_cnt: {alert_cnt}"
                    )
                except Exception as e:
                    logger.info("Skip headerm or non numeric org id %s", e)
                    continue
                if (
                    prev_env_org_policy is not None
                    and cur_env_org_policy != prev_env_org_policy
                ):
                    matched_rules += len(cached_matched_rules)
                    write_rules_to_csv(
                        local_output_file,
                        cached_matched_rules,
                        prev_env_org_policy,
                        write_header,
                    )
                    # only write the header one time
                    if write_header:
                        write_header = False
                    cached_matched_rules.clear()
                prev_env_org_policy = cur_env_org_policy
                # matched it before
                if proc_path in rule_path_match:
                    (
                        list_of_path_match,
                        num_visits,
                        num_path_tokens,
                    ) = rule_path_match[proc_path]
                else:
                    (
                        list_of_path_match,
                        num_visits,
                        num_path_tokens,
                    ) = self.find_path_in_tree(proc_path)
                    rule_path_match[proc_path] = (
                        list_of_path_match,
                        num_visits,
                        num_path_tokens,
                    )

                if len(list_of_path_match) > 0:
                    # For stats
                    num_nodes_visted_arr.append(num_visits)
                    num_matching_path_arr.append(len(list_of_path_match))
                    num_path_tokens_arr.append(num_path_tokens)
                for key in list_of_path_match:
                    path_match = key
                    if path_match is not None and path_match in self.rule_of:
                        if key in cached_matched_rules:
                            cached_matched_rules[key]["dev_cnt"] = max(
                                cached_matched_rules[key]["dev_cnt"], dev_cnt
                            )
                            cached_matched_rules[key]["alert_cnt"] += alert_cnt
                        else:
                            cached_matched_rules[key] = {
                                "dev_cnt": dev_cnt,
                                "alert_cnt": alert_cnt,
                                "alert_dev_cnt": 0,
                                "rule_sugg": self.rule_of[path_match],
                            }

        if len(cached_matched_rules) > 0:
            write_rules_to_csv(
                local_output_file,
                cached_matched_rules,
                cur_env_org_policy,
                write_header,
            )
            matched_rules += len(cached_matched_rules)
        return matched_rules
