# Native imports
import json
import time
import uuid
from collections.abc import MutableMapping

# 3rd party imports
import requests

from mycroft.common import settings


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


class Simpledict(MutableMapping):
    """ Dictionary based on MutableMapping

    :rtype : Simpledict
    """

    # noinspection PyMissingConstructor
    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(dict(*args, **kwargs))

    def __getitem__(self, key):
        return self._store[key]

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __repr__(self):
        return self._store.__repr__()


# Add your creds here
headers = Simpledict()

headers["X-Tic-User-Key"] = settings.TIC_USER_KEY
headers["X-Tic-User-Secret"] = settings.TIC_USER_SECRET
headers["Content-Type"] = "application/json"

url = "https://threatintel.bit9.com/api/v1/defense/reputation"

# Make a session for the defense API
session = requests.Session()
session.headers.update(headers)


def chunk(iterable, chunk_size):
    """
    Yields chunks of the specified size from the provided iterable.

    :param iterable: The item to iterate across.
    :param chunk_size: The size of the chunk to yield.
    :return: The chunks.
    """

    items = []

    for item in iterable:
        # Add elements until batch size is reached
        items.append(item)
        if len(items) != chunk_size:
            continue

        # Yield the item and reset the list
        yield items
        items = []

    # Yield any remaining items
    if len(items) > 0:
        yield items


def yield_reputation(sha256s):
    """
    Yields reputation results.

    :param sha256s: An iterable of SHA256s to obtain reputation for.
    :return: A list of reputation results.
    """

    # Wait 1s between request calls
    sleep_time = 1

    # Iterate over items in batches of 100
    for c in chunk(sha256s, 100):
        uq_sha256s = set(c)  # Dedupe
        n_attempts = 0

        # Poll the service until we get the data - do at most 10 attempts
        while len(uq_sha256s) > 0 and n_attempts < 10:
            # Build the body and make the request
            body = [{"sha256": sha256} for sha256 in uq_sha256s]
            response = session.post(url, data=json.dumps(body))

            # If transient errors occur, retry
            if response.status_code != 200:
                if response.status_code == 401:
                    raise (Exception("Invalid credentials"))
                time.sleep(sleep_time)
                n_attempts += 1
                continue

            # Request was good, figure out what results need to be resubmitted
            results = response.json()["results"]
            uq_sha256s = set()
            for result in results:
                if result["status"] not in {"Found", "NotFound"}:
                    uq_sha256s.add(result["hash"])
                else:
                    yield result
            if len(uq_sha256s) == 0:
                return
            n_attempts += 1
            time.sleep(sleep_time)

        # If we get here, then we failed to get reputation after 10 attempts
        raise (Exception("Failed to get reputation"))


def process_hashes(sha256s):
    """
    Get reputation for SHA2565s. Group by the category.

    :param sha256s: An iterable of SHA256s to obtain reputation for.
    :return: A dictionary mapping category to SHA256.
    """

    # Dictionary of category to SHA256
    results = {}

    for result in yield_reputation(sha256s):
        category = (
            "NOT_LISTED"
            if result["status"] == "NotFound"
            else result["reputation"]["category"]
        )
        found_results = results.get(category, set())
        found_results.add(result["hash"])
        if len(found_results) == 1:
            results[category] = found_results

    for category, uq_sha256s in results.items():
        print("Category: {}".format(category))
        for sha256 in uq_sha256s:
            print("--> {}".format(sha256))
        print("")
    return results


if __name__ == "__main__":
    process_hashes(
        [
            "d691ad30276f901e1921bcf4cd5e669c81938775e2e7657b4b2bb5ff66b57134",  # TRUSTED_WHITE_LIST
            (str(uuid.uuid4()) + str(uuid.uuid4())).replace("-", ""),  # NOT_LISTED
            "ccd93a7899d6fb29861f2cacf793950c5898249a1f007d804aa0df0a2a40a7ff",  # COMMON_WHITE_LIST
            "fc239ac15ee4dcf3a0dcb16b3c543c28cc0024c02093dfc2a127ac13b01c5ab4",  # KNOWN_MALWARE
            "fc392ccb71e281d0b8e27e9c45f9f75a43e0294c546a3490f7985e2ab24ed4c6",  # ADAPTIVE_WHITE_LIST
            "fd19368a5fc17b64bc84cfef872cb8fb39a63a2d3a53614470ceb13cd7302980",  # PUP
            "18490a92a717c7bc66ef548d239b6f2884c4382e39c9de9ab20204113dfe8539",  # SUSPECT_MALWARE
        ]
    )
