from enum import Enum

from marshmallow import Schema, fields, validate
from marshmallow_enum import EnumField


class DeviceGroup(Schema):
    """[Group of devices controlled using a single policy]
    Arguments:
        Schema {[type]} -- [description]
    """

    device_group_id = fields.Integer(
        required=True
    )  # Unique identifier for the device group
    org_id = fields.Integer(required=True)  # Organization id
    policy_control = fields.Boolean(
        required=True
    )  # Boolean indicator whether the device policy is controlled or not
    is_default = fields.Boolean(
        required=True
    )  # Boolean indicator whether this is the default group for the organization
    priority = fields.Int(
        validate=validate.Range(min=0, max=100)
    )  # Priority or importance of this group
    name = fields.Str(required=True)  # 'Device group name'
    description = fields.Str(
        validate=validate.Length(max=1024)
    )  # Device Group Description
    analytics_profile = fields.Str(
        validate=validate.Length(max=255)
    )  # 'Analytics profile associated with the device group
    analytics_version = fields.Str(
        validate=validate.Length(max=255)
    )  # 'Analytics version associated with the device group
    admin_contact_id = fields.Integer()  # 'Administrative contact for this device group
    policy_info = fields.Str(
        validate=validate.Length(max=4096)
    )  # 'Serialized Policy used to control devices associated with this device group
    policy_info_v2 = (
        fields.Str()
    )  # 'Serialized Policy used to control devices associated with this device group
    update_version = fields.Integer(
        required=True
    )  # 'Version counter used to check for concurrent updates
    priority_level = fields.Integer(required=True)


class DevicTypeEnum(Enum):
    Windows = 1
    Android = 2
    Mac = 3
    Ios = 4
    Other = 10
    # TODO add Linux


class DeviceType(Schema):
    # Device type - indicating whether it is a Windows, Android, Iphone.
    device_type_id = EnumField(
        DevicTypeEnum, required=True
    )  # Unique identifier for the device type
    name = fields.Str(required=True)  # Name of the device type


class Device(Schema):

    """
    Device - all the devices known to the system
    """

    device_id = fields.Integer(
        required=True
    )  # Unique numeric identifier for the device
    # device_type_id = fields.Integer(required=True) # Device type - windows, android etc.
    device_type_id = EnumField(
        DevicTypeEnum, required=True
    )  # Device type - windows, android etc.
    device_owner_id = fields.Integer(
        required=True
    )  # Device Owner - references device_owner entity
    device_guid = fields.Str(
        validate=validate.Length(max=255), required=True
    )  # Global unique identifer for the device generated on the device itself
    name = fields.Str(
        validate=validate.Length(max=1024)
    )  # Name of the device - could be machine name or some other identifier
    email = fields.Str(
        validate=validate.Length(max=1024)
    )  # Email id that the device is registered with
    email_type_id = fields.Integer()  # Email type id indicating work or personal email
    session_id = fields.Str(
        validate=validate.Length(max=4096)
    )  # Android device registration ID to google cloud messaging
    properties = fields.Str(
        validate=validate.Length(max=4096)
    )  # Key value pairs of device-type specific properties for the device
    assignedToId = fields.Integer()  # Id of the admin assigned to this device
    assignedToName = fields.Str(
        validate=validate.Length(max=255)
    )  # Name of the admin assigned to this device
    create_time = (
        fields.Integer()
    )  # Time the device was created in timestamp with time zone format
    update_version = fields.Integer(
        required=True
    )  # Version counter used to check for concurrent updates
    time_correction = (
        fields.Integer()
    )  # Time correction in milliseconds to apply to behavior event times from this device
    replica_device_id = (
        fields.Integer()
    )  # Id of mapped device in test org. All events for device will be mapped to this replica device id and rerun through analytics
    change_timestamps = fields.Str(
        validate=validate.Length(max=4096)
    )  # 'Timestamps used to track change events on the device


class DeviceGroupDevice(Schema):

    """
    device_group_device - Device that are part of a device group
    """

    device_group_id = fields.Integer(
        required=True
    )  # Unique identifier for the device group
    device_id = fields.Integer(required=True)  # Unique identifier for the device


# OrgReputationOverride
class ReputationRule(Schema):
    """
    An entry in OrgReputationOverride ES index that's used to override the reputation
    from the cloud API based on the the organizations wishes.
    """

    sourceType = fields.Str(required=True, validate=validate.OneOf(["APP", "FILE"]))
    orgId = fields.Int(required=True)
    env = fields.Str(required=True)
    csrAction = fields.Boolean(required=True)
    createdBy = fields.Str()
    value1 = fields.Str(required=True)
    value2 = fields.Str()
    createTime = fields.Integer()
    authListType = fields.Str(
        required=True, validate=validate.OneOf(["BLACK_LIST", "WHITE_LIST"])
    )
    lastUpdatedTime = fields.Integer(required=True)
    fileId = fields.Str()
    entryId = fields.Str()
    lastModification = fields.Str(
        required=True, validate=validate.OneOf(["UPDATE", "DELETE", "ADD"])
    )
    # description = fields.Str(required=True)
    type = fields.Str(
        required=True, validate=validate.OneOf(["CERT", "IT_TOOL", "SHA256"])
    )
    updateVersion = fields.Int()


# CBD sensor rules Application
class Application(Schema):
    type = fields.Str(
        required=True, validate=validate.OneOf(["REPUTATION", "NAME_PATH"])
    )
    value = fields.Str(required=True)


# aka permission rule
class PolicyRule(Schema):
    """[summary]
    """

    application = fields.Nested(Application)
    operation = fields.Str(
        required=True,
        validate=validate.OneOf(
            [
                "BUFFER_OVERFLOW",
                "BYPASS_ALL",
                "BYPASS_API",
                "CODE_INJECTION",
                "INVOKE_CMD_INTERPRETER",
                "INVOKE_SCRIPT",
                "MEMORY_SCRAPE",
                "NETWORK",
                "PACKED",
                "POL_INVOKE_NOT_TRUSTED",
                "RANSOM",
                "RUN",
            ]
        ),
    )
    action = fields.Str(required=True, validate=validate.OneOf(["ALLOW", "IGNORE"]),)


class RepRule(Schema):
    rule = fields.Nested(ReputationRule)


class PemRule(Schema):
    rule = fields.Nested(PolicyRule)
