#!/python3
import logging
import os
import re
import traceback

import boto3

from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


def parse_parquet_filename(filename: str):
    regex = r"part-\d\d\d\d\d-(.*)-[a-z0-9]+\.gz.parquet"
    match = re.findall(regex, filename)
    if match:
        return match[0]
    else:
        return None


def copy(s3, source_bucket, source_key, dest_bucket, dest_key):
    copy_source = {"Bucket": source_bucket, "Key": f"{source_key}"}
    logger.debug(
        f"going to copy s3://{source_bucket}/{source_key} to s3://{dest_bucket}/{dest_key}"
    )
    s3.meta.client.copy(copy_source, dest_bucket, dest_key)


def cleanup(s3_client, output_bucket: str, s3_prefix: str, last_extraction_id: str):
    """
    Remove all s3 files under the specific prefix that don't meet the extraction_id search criterion.
    """
    bucket = s3_client.Bucket(output_bucket)

    # remove data
    res = bucket.meta.client.list_objects_v2(
        Bucket=output_bucket,
        Delimiter="/",
        # EncodingType='url',
        Prefix=s3_prefix,
    )
    list_for_deletion = []
    list_for_keep = []
    if "Contents" in res:
        for key in res["Contents"]:
            if last_extraction_id in key["Key"]:
                list_for_keep.append(key["Key"])
            else:
                list_for_deletion.append(key["Key"])
    logger.debug(f"Clean up: delete: {list_for_deletion}")
    logger.debug(f"Clean up: keep: [{list_for_keep}]")
    for item in list_for_deletion:
        bucket.meta.client.delete_object(Bucket=output_bucket, Key=item)


def move_last_export(**kwargs):
    try:
        logging.debug("Start move_last_export")
        region = kwargs["region"]
        env = kwargs["env"]
        legacy_env = kwargs["legacy_env"]
        bucket_name = kwargs["bucket"]
        logging.debug(
            "Input params bucket = %s, region = %s, env = %s", bucket_name, region, env
        )
        output_prefix_policy = f"policy/device_group/env={env}"
        output_prefix_device_mapping = f"policy/device_group_device/env={env}"

        s3 = boto3.resource("s3", region_name=region)
        bucket = s3.Bucket(bucket_name)
        logging.debug("s3 bucket = %s", bucket_name)
        # get all files from the source bucket.
        result = bucket.meta.client.list_objects(Bucket=bucket.name, Delimiter="/")
        logging.debug("list all files in source bucket = %s", result)
        # Identifies the latest export that is created by the rds_snapshot_export DAG using the file prefix.
        last_export_prefix = [x for x in result.get("CommonPrefixes")]
        logging.debug("last export prefix = %s", last_export_prefix)
        # identify the last export date
        if legacy_env:
            last_export = [
                d for d in last_export_prefix if re.match(legacy_env, d["Prefix"])
            ][-1]["Prefix"]
            logging.debug(f"DEBUG: last export: {last_export}")
            db = f"{legacy_env}-psc-db-aurora"
            last_export_date = last_export.replace(db, "")
            logger.debug(f"Found last export on {last_export_date}\n")
            start_extraction(
                s3,
                bucket,
                last_export,
                legacy_env,
                "public.device_group",
                output_prefix_policy,
            )
            start_extraction(
                s3,
                bucket,
                last_export,
                legacy_env,
                "public.device_group_device",
                output_prefix_device_mapping,
            )
        else:
            last_export = [d for d in last_export_prefix if re.match(env, d["Prefix"])][
                -1
            ]["Prefix"]
            logging.debug(f"DEBUG: last export: {last_export}")
            db = f"{legacy_env}-cbd-aurora-postgres"
            last_export_date = last_export.replace(db, "")
            logger.debug(f"Found last export on {last_export_date}\n")
            start_extraction(
                s3,
                bucket,
                last_export,
                "pscdbaurora",
                "public.device_group",
                output_prefix_policy,
            )
            start_extraction(
                s3,
                bucket,
                last_export,
                "pscdbaurora",
                "public.device_group_device",
                output_prefix_device_mapping,
            )
    except Exception:
        logging.debug(traceback.format_exc())
        raise Exception("Failed to move_data")


def start_extraction(s3, bucket, db, table, field, output_prefix):
    """
    :param s3: File to upload
    :param bucket
    :param db Database name with snapshot export time. Ex: dev01-psc-db-aurora2023-10-06-22-05-04/
    :param table Table name.
    :param field Target field to be extracted. Ex: public.device_group
    :param output_prefix Bucket prefix for
    """
    # moving data for device_group_device (aka device_policy_mapping) tables
    # MOSAIC2: hard coded the prefix to database name. In Mosaic2, a folder name is created based on database name.
    # Legacy env: dev01-psc-db-aurora2023-09-29-17-34-19/dev01/public.device_group/1/
    # UkPop: int1usw2-cbd-aurora-postgres2023-10-05-22-05-29/pscdbaurora/public.device_group/1
    input_prefix = f"{db}{table}/{field}/"
    logger.debug(
        f"Start extracting [{input_prefix}] to [{bucket.name}] with prefix {output_prefix}"
    )
    result = bucket.meta.client.list_objects(
        Bucket=bucket.name,
        Delimiter="/",
        Prefix=f"{input_prefix}",
    )

    if result.get("Contents") is None:
        logger.debug("Looking under folder 1/ for the parquet file")
        result = bucket.meta.client.list_objects(
            Bucket=bucket.name,
            Delimiter="/",
            Prefix=f"{input_prefix}1/",
        )

    # we're calling parse_parquet_filename to get the last extraction id so we can remove everything except for that
    last_extraction_id = ""
    for k in result.get("Contents"):
        if k.get("Key").endswith("parquet"):
            logger.debug(k.get("Key"), "\n")
            filename = os.path.basename(k.get("Key"))
            if last_extraction_id == "":
                last_extraction_id = parse_parquet_filename(filename)
            copy(
                s3,
                bucket.name,
                k.get("Key"),
                bucket.name,
                f"{output_prefix}/{filename}",
            )
    cleanup(s3, bucket.name, f"{output_prefix}/", last_extraction_id)
